/******************************************************************
 *                       PINOCCHI0  V3.0                         *
 *  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
 *****************************************************************
 
 This code was written by
 Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
 Copyright (C) 2013
 
 web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*

  The code contained in this file has been adapted from the 2LPTic code
  by Roman Scoccimarro, downloadable from:

  http://cosmo.nyu.edu/roman/2LPT/

  The 2LPTic code is a 2LPT extension of the N-GenIC code by V. Springel,
  downloadable from:

  http://www.gadgetcode.org/right.html#ICcode

  and distributed under GNU/GPL license.  The part of the 2LPTic code 
  we use here is entirely contained in N-GenIC.

*/



#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>
#include <string.h>
#include <gsl/gsl_rng.h>

#define  PI          3.14159265358979323846 
#define  GRAVITY     6.672e-8
#define  HUBBLE      3.2407789e-18   /* in h/sec */

#define BLENGTH 300

#define NYQUIST 1.0  	/* filtering at the Nyquist frequency, standard pinocchio: 0.8 */

//#define DEBUG

#ifndef DO_NOT_USE_GENIC

int    density_in_kspace(int*);
void   set_units(void);
double fnl(double x);
void   read_power_table(void);
double PowerSpec(double kmag);
double PowerSpec_Efstathiou(double k);
double PowerSpec_EH(double k);
double PowerSpec_Tabulated(double k);
void   initialize_powerspectrum(void);
double sigma2_int(double k);
double TopHatSigma2(double R);
int    compare_logk(const void *a, const void *b);
double tk_eh(double k);
double transf_eh_(void);

int    WhichSpectrum;
FILE   *FdTmp, *FdTmpInput;
int    Nmesh, Nsample;
char   FileWithInputSpectrum[500];
double Box;
int    Seed;
char   OutputDir[100], FileBase[100];
int    ThisTask, NTask;
int    Local_nx, Local_x_start;

double UnitTime_in_s, UnitLength_in_cm, UnitMass_in_g, UnitVelocity_in_cm_per_s;
double InputSpectrum_UnitLength_in_cm;
double G, Hubble;
double RhoCrit;

double Omega, OmegaLambda, Sigma8;
double OmegaBaryon, HubbleParam, DEw0, DEw1, StartingzForPLC;
double PrimordialIndex;
double ShapeGamma;

double WDM_PartMass_in_kev;

extern struct
{
  double *vector;
} comm_genic_;

extern struct
{
  double p,kappa;
} comm_wrap_;

int genic_(int *VectorLength, int *Local_nx_, int *Local_x_start_, int *ThisTask_, int *NTask_, int *init)
{

  extern struct
  {
    double Omega0, OmegaLambda, Hubble100, Sigma8, PowerSpectrum, 
      BoxSize, OmegaBaryon, DEw0, DEw1, StartingzForPLC,
      InputSpectrum_UnitLength_in_cm, WDM_PartMass_in_kev;
    int GridSize[3],RandomSeed,ResamplingRate;
    int WriteFmax, WriteVmax, WriteRmax, WriteDensity, WriteDisplacements,
      CatalogInAscii, DoNotFragment, OutputInH100,BoxInH100;
    char Paramfile[BLENGTH],RunFlag[BLENGTH],OutputList[BLENGTH],
      FileWithInputSpectrum[BLENGTH];
    char TabulatedEoSfile[BLENGTH];  //, ReadDensityField[BLENGTH], StartFromSavedFmax[BLENGTH];
  } params_;

  int j;
  char buf[BLENGTH];

  if (*init)
    {
      Local_nx=*Local_nx_;
      Local_x_start=*Local_x_start_;

      Nmesh           = params_.GridSize[0];
      Nsample         = params_.GridSize[0];
      Omega           = params_.Omega0;
      OmegaLambda     = params_.OmegaLambda;
      Sigma8          = params_.Sigma8;
      OmegaBaryon     = params_.OmegaBaryon;
      DEw0            = params_.DEw0;
      DEw1            = params_.DEw1;
      StartingzForPLC = params_.StartingzForPLC;
      HubbleParam     = params_.Hubble100;
      PrimordialIndex = params_.PowerSpectrum;
      Seed            = params_.RandomSeed;
      if (params_.BoxInH100)
	Box             = params_.BoxSize;
      else
	Box             = params_.BoxSize*params_.Hubble100;

      WDM_PartMass_in_kev = params_.WDM_PartMass_in_kev;
      if (params_.InputSpectrum_UnitLength_in_cm>0)
	InputSpectrum_UnitLength_in_cm = params_.InputSpectrum_UnitLength_in_cm;
      else
	InputSpectrum_UnitLength_in_cm = 3.085678e24;

      for(j=0;j<300;j++)
	if (params_.FileWithInputSpectrum[j]==' ')
	  {
	    buf[j]='\0';
	    break;
	  }
	else
	  buf[j]=params_.FileWithInputSpectrum[j];
      strcpy(FileWithInputSpectrum,buf);
  

      /* these variables are set to predefined values */
      if (strcmp(FileWithInputSpectrum,"no"))
	WhichSpectrum=2;
      else
	WhichSpectrum=1;

      strcpy(OutputDir,".");
      strcpy(FileBase,"GenIC");
      ShapeGamma=0.21;  /* Efstathiou et al's parameterization is not used */
      //      Redshift=0;       /* this is used to output the power spectrum */

      UnitLength_in_cm         = 3.085678e24;
      UnitMass_in_g            = 1.989e43;
      UnitVelocity_in_cm_per_s = 1e5;

      ThisTask = *ThisTask_;
      NTask    = *NTask_;


      set_units();

      initialize_powerspectrum();
      
      //      print_spec();

    }
  else
    {

      if (density_in_kspace(VectorLength))
	return 1;
    }

  return 0;
}


int density_in_kspace(int *VectorLength)
{
  gsl_rng *random_generator;
  int i, j, k, ii, jj;
  double fac;
  double kvec[3], kmag, kmag2, p_of_k;
  double delta, phase, ampl;
  unsigned int *seedtable;
  //int index, iproc;

  fac = pow(2 * PI / Box, 1.5);

  /* This is the initialization of the random number generator */
  random_generator = gsl_rng_alloc(gsl_rng_ranlxd1);

  gsl_rng_set(random_generator, Seed);

  if(!(seedtable = malloc(Nmesh * Nmesh * sizeof(unsigned int))))
    return 4;

  for(i = 0; i < Nmesh / 2; i++)
    {
      for(j = 0; j < i; j++)
	seedtable[i * Nmesh + j] = 0x7fffffff * gsl_rng_uniform(random_generator);

      for(j = 0; j < i + 1; j++)
	seedtable[j * Nmesh + i] = 0x7fffffff * gsl_rng_uniform(random_generator);

      for(j = 0; j < i; j++)
	seedtable[(Nmesh - 1 - i) * Nmesh + j] = 0x7fffffff * gsl_rng_uniform(random_generator);

      for(j = 0; j < i + 1; j++)
	seedtable[(Nmesh - 1 - j) * Nmesh + i] = 0x7fffffff * gsl_rng_uniform(random_generator);

      for(j = 0; j < i; j++)
	seedtable[i * Nmesh + (Nmesh - 1 - j)] = 0x7fffffff * gsl_rng_uniform(random_generator);

      for(j = 0; j < i + 1; j++)
	seedtable[j * Nmesh + (Nmesh - 1 - i)] = 0x7fffffff * gsl_rng_uniform(random_generator);

      for(j = 0; j < i; j++)
	seedtable[(Nmesh - 1 - i) * Nmesh + (Nmesh - 1 - j)] = 0x7fffffff * gsl_rng_uniform(random_generator);

      for(j = 0; j < i + 1; j++)
	seedtable[(Nmesh - 1 - j) * Nmesh + (Nmesh - 1 - i)] = 0x7fffffff * gsl_rng_uniform(random_generator);
    }


  /* the vector is initialized to zero */
  for (i = 0; i < *VectorLength; i++)
    comm_genic_.vector[i]=0.;

  /* this is the main loop on grid modes */
  for(i = 0; i < Nmesh; i++)
    {
      ii = Nmesh - i;
      if(ii == Nmesh)
	ii = 0;
      if((i >= Local_x_start && i < (Local_x_start + Local_nx)) ||
	 (ii >= Local_x_start && ii < (Local_x_start + Local_nx)))
	{
	  for(j = 0; j < Nmesh; j++)
	    {
	      gsl_rng_set(random_generator, seedtable[i * Nmesh + j]);

	      for(k = 0; k < Nmesh / 2; k++)
		{
		  phase = gsl_rng_uniform(random_generator) * 2 * PI;
		  do
		    ampl = gsl_rng_uniform(random_generator);
		  while(ampl == 0);

		  if(i == Nmesh / 2 || j == Nmesh / 2 || k == Nmesh / 2)
		    continue;
		  if(i == 0 && j == 0 && k == 0)
		    continue;

		  if(i < Nmesh / 2)
		    kvec[0] = i * 2 * PI / Box;
		  else
		    kvec[0] = -(Nmesh - i) * 2 * PI / Box;

		  if(j < Nmesh / 2)
		    kvec[1] = j * 2 * PI / Box;
		  else
		    kvec[1] = -(Nmesh - j) * 2 * PI / Box;

		  if(k < Nmesh / 2)
		    kvec[2] = k * 2 * PI / Box;
		  else
		    kvec[2] = -(Nmesh - k) * 2 * PI / Box;

		  kmag2 = kvec[0] * kvec[0] + kvec[1] * kvec[1] + kvec[2] * kvec[2];
		  kmag = sqrt(kmag2);

		  /* if(kmag * Box / (2 * PI) > Nsample / 2)        select a sphere in k-space */
		  if(kmag * Box / (2 * PI) > NYQUIST * Nsample / 2)
		    continue;

		  /*
		  else
		    {
		      if(fabs(kvec[0]) * Box / (2 * PI) > Nsample / 2)
			continue;
		      if(fabs(kvec[1]) * Box / (2 * PI) > Nsample / 2)
			continue;
		      if(fabs(kvec[2]) * Box / (2 * PI) > Nsample / 2)
			continue;
		    }
		  */

		  p_of_k = PowerSpec(kmag);

		  p_of_k *= -log(ampl);

		  delta = fac * sqrt(p_of_k);

		  if(k > 0)
		    {
		      if(i >= Local_x_start && i < (Local_x_start + Local_nx))
			{

			  comm_genic_.vector[2*(((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k)  ] =  delta * cos(phase);
			  comm_genic_.vector[2*(((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k)+1] =  delta * sin(phase);
			  
			}

		      /*  ORIGINAL
			for(axes = 0; axes < 3; axes++)
			{
			cdisp[axes][((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k].re =
			-kvec[axes] / kmag2 * delta * sin(phase);
			cdisp[axes][((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k].im =
			kvec[axes] / kmag2 * delta * cos(phase);
			}
		      */
		    }
		  else	/* k=0 plane needs special treatment */
		    {
		      if(i == 0)
			{
			  if(j >= Nmesh / 2)
			    continue;
			  else
			    {
			      if(i >= Local_x_start && i < (Local_x_start + Local_nx))
				{
				  jj = Nmesh - j;	/* note: j!=0 surely holds at this point */
				  
				  comm_genic_.vector[2*(((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k)  ] = delta * cos(phase);
				  comm_genic_.vector[2*(((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k)+1] = delta * sin(phase);

				  comm_genic_.vector[2*(((i - Local_x_start) * Nmesh + jj) * (Nmesh / 2 + 1) + k)  ] = delta * cos(phase);
				  comm_genic_.vector[2*(((i - Local_x_start) * Nmesh + jj) * (Nmesh / 2 + 1) + k)+1] = -delta * sin(phase);

				  /*  ORIGINAL
				    for(axes = 0; axes < 3; axes++)
				    {
				    cdisp[axes][((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k].re =
				    -kvec[axes] / kmag2 * delta * sin(phase);
				    cdisp[axes][((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k].im =
				    kvec[axes] / kmag2 * delta * cos(phase);
					  
				    cdisp[axes][((i - Local_x_start) * Nmesh + jj) * (Nmesh / 2 + 1) + k].re =
				    -kvec[axes] / kmag2 * delta * sin(phase);
				    cdisp[axes][((i - Local_x_start) * Nmesh + jj) * (Nmesh / 2 + 1) + k].im =
				    -kvec[axes] / kmag2 * delta * cos(phase);
				    }
				  */
				}
			    }
			}
		      else	/* here comes i!=0 : conjugate can be on other processor! */
			{
			  if(i >= Nmesh / 2)
			    continue;
			  else
			    {
			      ii = Nmesh - i;
			      if(ii == Nmesh)
				ii = 0;
			      jj = Nmesh - j;
			      if(jj == Nmesh)
				jj = 0;

			      if(i >= Local_x_start && i < (Local_x_start + Local_nx))
				{

				  comm_genic_.vector[2*(((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k)  ] = delta * cos(phase);
				  comm_genic_.vector[2*(((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k)+1] = delta * sin(phase);

				}

			      /*  ORIGINAL
				for(axes = 0; axes < 3; axes++)
				{
				cdisp[axes][((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k].re =
				-kvec[axes] / kmag2 * delta * sin(phase);
				cdisp[axes][((i - Local_x_start) * Nmesh + j) * (Nmesh / 2 + 1) + k].im =
				kvec[axes] / kmag2 * delta * cos(phase);
				}
			      */

			      if(ii >= Local_x_start && ii < (Local_x_start + Local_nx))
				{
				  
				  comm_genic_.vector[2*(((ii - Local_x_start) * Nmesh + jj) * (Nmesh / 2 + 1) + k)  ] = delta * cos(phase);
				  comm_genic_.vector[2*(((ii - Local_x_start) * Nmesh + jj) * (Nmesh / 2 + 1) + k)+1] = -delta * sin(phase);
				  
				}

			      /*  ORIGINAL
				for(axes = 0; axes < 3; axes++)
				{
				cdisp[axes][((ii - Local_x_start) * Nmesh + jj) * (Nmesh / 2 + 1) +
				k].re = -kvec[axes] / kmag2 * delta * sin(phase);
				cdisp[axes][((ii - Local_x_start) * Nmesh + jj) * (Nmesh / 2 + 1) +
				k].im = -kvec[axes] / kmag2 * delta * cos(phase);
				}
			      */
			    }
			}
		    }
		}
	    }
	}
    }

  gsl_rng_free(random_generator);

  /* This is needed to achieve proper normalization for pinocchio */

  fac=pow((double)Nmesh, 3.0);
  for (i=0; i<*VectorLength; i++)
    comm_genic_.vector[i] *= fac;

  return 0;
}

void set_units(void)		/* ... set some units */
{
  UnitTime_in_s = UnitLength_in_cm / UnitVelocity_in_cm_per_s;

  G = GRAVITY / pow(UnitLength_in_cm, 3) * UnitMass_in_g * pow(UnitTime_in_s, 2);
  Hubble = HUBBLE * UnitTime_in_s;
}


static double R8;
static double r_tophat;

static double AA, BB, CC;
static double nu;
static double Norm;


static int NPowerTable;

static struct pow_table
{
  double logk, logD;
}
 *PowerTable;


int power_wrap_(double *k)
{

  comm_wrap_.p=PowerSpec(*k/HubbleParam)/pow(HubbleParam,3.0)*4.*PI;

  return 1;
}


double PowerSpec(double k)
{
  double power, alpha, Tf;

  switch (WhichSpectrum)
    {
    case 1:
      power = PowerSpec_EH(k);
      break;

    case 2:
      power = PowerSpec_Tabulated(k);
      break;

    default:
      power = PowerSpec_Efstathiou(k);
      break;
    }


  if(WDM_PartMass_in_kev > 0.)
    {
      /* Eqn. (A9) in Bode, Ostriker & Turok (2001), assuming gX=1.5  */
      alpha =
	0.048 * pow((Omega - OmegaBaryon) / 0.4, 0.15) * pow(HubbleParam / 0.65,
							     1.3) * pow(1.0 / WDM_PartMass_in_kev, 1.15);
      Tf = pow(1 + pow(alpha * k * (3.085678e24 / UnitLength_in_cm), 2 * 1.2), -5.0 / 1.2);
      power *= Tf * Tf;
    }

  return power;
}


void read_power_table(void)
{
  FILE *fd;
  char buf[500];
  double k, p;


  strcpy(buf, FileWithInputSpectrum);

  if(!(fd = fopen(buf, "r")))
    {
      printf("can't read input spectrum in file '%s' on task %d\n", buf, ThisTask);
      exit(17);
      //FatalError(17);
    }

  NPowerTable = 0;
  do
    {
      if(fscanf(fd, " %lg %lg ", &k, &p) == 2)
	NPowerTable++;
      else
	break;
    }
  while(1);

  fclose(fd);


  if(ThisTask == 0)
    {
      printf("found %d pairs of values in input spectrum table\n", NPowerTable);
      fflush(stdout);
    }


  PowerTable = malloc(NPowerTable * sizeof(struct pow_table));

  strcpy(buf, FileWithInputSpectrum);

  if(!(fd = fopen(buf, "r")))
    {
      printf("can't read input spectrum in file '%s' on task %d\n", buf, ThisTask);
      exit(18);
      //FatalError(18);
    }

  NPowerTable = 0;
  do
    {
      if(fscanf(fd, " %lg %lg ", &k, &p) == 2)
	{
	  PowerTable[NPowerTable].logk = k;
	  PowerTable[NPowerTable].logD = p;
	  NPowerTable++;
	}
      else
	break;
    }
  while(1);

  fclose(fd);

  qsort(PowerTable, NPowerTable, sizeof(struct pow_table), compare_logk);
}

int compare_logk(const void *a, const void *b)
{
  if(((struct pow_table *) a)->logk < (((struct pow_table *) b)->logk))
    return -1;

  if(((struct pow_table *) a)->logk > (((struct pow_table *) b)->logk))
    return +1;

  return 0;
}

void initialize_powerspectrum(void)
{
  double res;

  //InitTime = 1 / (1 + Redshift);

  AA = 6.4 / ShapeGamma * (3.085678e24 / UnitLength_in_cm);
  BB = 3.0 / ShapeGamma * (3.085678e24 / UnitLength_in_cm);
  CC = 1.7 / ShapeGamma * (3.085678e24 / UnitLength_in_cm);
  nu = 1.13;

  R8 = 8 * (3.085678e24 / UnitLength_in_cm);	/* 8 Mpc/h */


  if(WhichSpectrum == 2)
    read_power_table();

  Norm = 1.0;
  res = TopHatSigma2(R8);

  if(ThisTask == 0 && WhichSpectrum == 2)
    printf("\nNormalization of spectrum in file:  Sigma8 = %g\n", sqrt(res));

  Norm = Sigma8 * Sigma8 / res;

  if(ThisTask == 0 && WhichSpectrum == 2)
    printf("Normalization adjusted to  Sigma8=%g   (Normfac=%g)\n\n", Sigma8, Norm);

}

double PowerSpec_Tabulated(double k)
{
  double logk, logD, P, kold, u, dlogk, Delta2;
  int binlow, binhigh, binmid;

  kold = k;

  k *= (InputSpectrum_UnitLength_in_cm / UnitLength_in_cm);	/* convert to h/Mpc */

  logk = log10(k);

  if(logk < PowerTable[0].logk || logk > PowerTable[NPowerTable - 1].logk)
    return 0;

  binlow = 0;
  binhigh = NPowerTable - 1;

  while(binhigh - binlow > 1)
    {
      binmid = (binhigh + binlow) / 2;
      if(logk < PowerTable[binmid].logk)
	binhigh = binmid;
      else
	binlow = binmid;
    }

  dlogk = PowerTable[binhigh].logk - PowerTable[binlow].logk;

  if(dlogk == 0)
    exit(777);
  //FatalError(777);

  u = (logk - PowerTable[binlow].logk) / dlogk;

  logD = (1 - u) * PowerTable[binlow].logD + u * PowerTable[binhigh].logD;

  Delta2 = pow(10.0, logD);

  P = Norm * Delta2 / (4 * M_PI * kold * kold * kold);

  return P;
}

double PowerSpec_Efstathiou(double k)
{
  return Norm * pow(k,PrimordialIndex) / pow(1 + pow(AA * k + pow(BB * k, 1.5) + CC * CC * k * k, nu), 2 / nu);
}


double PowerSpec_EH(double k)	/* Eisenstein & Hu */
{

  comm_wrap_.kappa=k * 3.085678e24 / UnitLength_in_cm * HubbleParam;
  double transf=transf_eh_();
  //  printf(" aaaaa %f %g %g\n",k,transf,tk_eh(k));
  return Norm * pow(k,PrimordialIndex) * pow(transf, 2);

  //  return Norm * k * pow(tk_eh(k), 2);
}




double tk_eh(double k)		/* from Martin White */
{
  double q, theta, ommh2, a, s, gamma, L0, C0;
  double tmp;
  double omegam, ombh2, hubble;

  /* other input parameters */
  hubble = HubbleParam;

  omegam = Omega;
  ombh2 = OmegaBaryon * HubbleParam * HubbleParam;

  /*
  if(OmegaBaryon == 0)
  ombh2 = 0.04 * HubbleParam * HubbleParam; */

  k *= (3.085678e24 / UnitLength_in_cm);	/* convert to h/Mpc */

  theta = 2.728 / 2.7;
  ommh2 = omegam * hubble * hubble;
  s = 44.5 * log(9.83 / ommh2) / sqrt(1. + 10. * exp(0.75 * log(ombh2))) * hubble;
  a = 1. - 0.328 * log(431. * ommh2) * ombh2 / ommh2
    + 0.380 * log(22.3 * ommh2) * (ombh2 / ommh2) * (ombh2 / ommh2);
  gamma = a + (1. - a) / (1. + exp(4 * log(0.43 * k * s)));
  gamma *= omegam * hubble;
  q = k * theta * theta / gamma;
  L0 = log(2. * exp(1.) + 1.8 * q);
  C0 = 14.2 + 731. / (1. + 62.5 * q);
  tmp = L0 / (L0 + C0 * q * q);

  return (tmp);
}



double romberg_int(double (*func)(double), double a, double b);

double TopHatSigma2(double R)
{
  r_tophat = R;

  return romberg_int(sigma2_int, 0., 500.0 * 1 / R);	/* note: 500/R is here chosen as 
						   integration boundary (infinity) */
}


double sigma2_int(double k)
{
  double kr, kr3, kr2, w, x;

  kr = r_tophat * k;
  kr2 = kr * kr;
  kr3 = kr2 * kr;

  if(kr < 1e-8)
    return 0;

  w = 3 * (sin(kr) / kr3 - cos(kr) / kr2);
  x = 4 * PI * k * k * w * w * PowerSpec(k);

  return x;
}

#define EPS 1.0e-8
#define JMAX 40
#define JMAXP (JMAX+1)
#define K 5

void interp(double *,double *, int, double, double *, double *);
double trapezoidal(double (*func)(double), double, double, int);

double romberg_int(double (*func)(double), double a, double b)
{
  double ss,dss;
  double s[JMAXP],h[JMAXP+1];
  int j;

  h[1]=1.0;
  for (j=1;j<=JMAX;j++)
    {
      s[j]=trapezoidal(func,a,b,j);
      if (j >= K)
	{
	  interp(&h[j-K],&s[j-K],K,0.0,&ss,&dss);
	  if (fabs(dss) <= EPS*fabs(ss))
	    return ss;
	}
      h[j+1]=0.25*h[j];
    }

  return 0.0;
}

#undef EPS
#undef JMAX
#undef JMAXP
#undef K

double trapezoidal(double (*func)(double), double a, double b, int n)
{
  double x,tnm,sum,del;
  static double s;
  int it,j;

  if (n == 1) 
    return (s=0.5*(b-a)*((*func)(a)+(*func)(b)));
  else
    {
      for (it=1,j=1;j<n-1;j++)
	it <<= 1;
      tnm=it;
      del=(b-a)/tnm;
      x=a+0.5*del;
      for (sum=0.0,j=1;j<=it;j++,x+=del)
	sum += (*func)(x);
      s=0.5*(s+(b-a)*sum/tnm);
      return s;
    }
}

void interp(double xa[], double ya[], int n, double x, double *y, double *dy)
{
  int i,m,ns=1;
  double den,dif,dift,ho,hp,w;
  double *c,*d;

  dif=fabs(x-xa[1]);
  c=(double *)malloc( (n+1)*sizeof(double) );
  d=(double *)malloc( (n+1)*sizeof(double) );

  for (i=1;i<=n;i++)
    {
      if ( (dift=fabs(x-xa[i])) < dif)
	{
	  ns=i;
	  dif=dift;
	}
      c[i]=ya[i];
      d[i]=ya[i];
    }
  *y=ya[ns--];
  for (m=1;m<n;m++)
    {
      for (i=1;i<=n-m;i++)
	{
	  ho=xa[i]-x;
	  hp=xa[i+m]-x;
	  w=c[i+1]-d[i];
	  if ( (den=ho-hp) == 0.0) printf("Error in routine interp\n");
	  den=w/den;
	  d[i]=hp*den;
	  c[i]=ho*den;
	}
      *y += (*dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
    }
  free(d);
  free(c);
}


#endif
