/******************************************************************
 *                       PINOCCHI0  V3.0                         *
 *  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
 *****************************************************************
 
 This code was written by
 Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
 Copyright (C) 2013
 
 web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <mpi.h>

#define BLENGTH 300
#define FILELENGTH 300
#define DOUBLE 1
#define STRING 2
#define INT 3
#define LOGICAL 4
#define INT3 5
#define INT_SKIP 99
#define MAXTAGS 50

#define MAXOUTPUTS 1000

void space_pad(char*);

int read_parameter_file_(int *proc_id)
{
  FILE *fd;
  char buf[BLENGTH], buf1[BLENGTH], buf2[BLENGTH], buf3[BLENGTH], buf4[BLENGTH];
  int i, j, nt, number_of_fields;
  int id[MAXTAGS];
  void *addr[MAXTAGS];
  char tag[MAXTAGS][FILELENGTH];
  char *ufo;
  double z;

  extern struct
  {
    double Omega0, OmegaLambda, Hubble100, Sigma8, PowerSpectrum, 
      BoxSize, OmegaBaryon, DEw0, DEw1, StartingzForPLC;
    double  InputSpectrum_UnitLength_in_cm, WDM_PartMass_in_kev;
    int GridSize[3],RandomSeed,ResamplingRate;
    int WriteFmax, WriteVmax, WriteRmax, WriteDensity, WriteDisplacements,
      CatalogInAscii, DoNotFragment, OutputInH100, BoxInH100;
    char Paramfile[BLENGTH],RunFlag[BLENGTH],OutputList[BLENGTH];
    char FileWithInputSpectrum[BLENGTH];
    char TabulatedEoSfile[BLENGTH];
    //char ReadDensityField[BLENGTH],StartFromSavedFmax[BLENGTH];
  } params_;

  extern struct
  {
    double zout[MAXOUTPUTS];
    int nout;
  } outputs_;

  if (*proc_id==0)
    {

      /* fixes fortran string */
      for(j=0;j<300;j++)
	if (params_.Paramfile[j]==' ')
	  {
	    buf[j]='\0';
	    break;
	  }
	else
	  buf[j]=params_.Paramfile[j];

      strcpy(params_.Paramfile,buf);


      printf("%s\n",params_.Paramfile);
      fflush(stdout);



      nt=0;

      /* list of requested parameters */
      strcpy(tag[nt], "RunFlag");
      addr[nt] = params_.RunFlag;
      id[nt++] = STRING;

      strcpy(tag[nt], "OutputList");
      addr[nt] = params_.OutputList;
      id[nt++] = STRING;

      /*
      strcpy(tag[nt], "ReadDensityField");
      addr[nt] = params_.ReadDensityField;
      id[nt++] = STRING;
      */

      strcpy(tag[nt], "Omega0");
      addr[nt] = &params_.Omega0;
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "OmegaLambda");
      addr[nt] = &params_.OmegaLambda;
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "PowerSpectrum");
      addr[nt] = &params_.PowerSpectrum;
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "Sigma8");
      addr[nt] = &params_.Sigma8;
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "Hubble100");
      addr[nt] = &params_.Hubble100;
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "OmegaBaryon");
      addr[nt] = &params_.OmegaBaryon;
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "DEw0");
      addr[nt] = &params_.DEw0;
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "DEw1");
      addr[nt] = &params_.DEw1;
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "StartingzForPLC");
      addr[nt] = &params_.StartingzForPLC;
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "BoxSize");
      addr[nt] = &params_.BoxSize;
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "BoxInH100");
      addr[nt] = &params_.BoxInH100;
      id[nt++] = LOGICAL;

      strcpy(tag[nt], "RandomSeed");
      addr[nt] = &(params_.RandomSeed);
      id[nt++] = INT;

      strcpy(tag[nt], "GridSize");
      addr[nt] = &(params_.GridSize[0]);
#ifdef NON_CUBIC
      id[nt++] = INT3;
#else
      id[nt++] = INT;
#endif

      strcpy(tag[nt], "WriteFmax");
      addr[nt] = &params_.WriteFmax;
      id[nt++] = LOGICAL;

      strcpy(tag[nt], "WriteVmax");
      addr[nt] = &params_.WriteVmax;
      id[nt++] = LOGICAL;

      strcpy(tag[nt], "WriteRmax");
      addr[nt] = &params_.WriteRmax;
      id[nt++] = LOGICAL;

      strcpy(tag[nt], "WriteDensity");
      addr[nt] = &params_.WriteDensity;
      id[nt++] = LOGICAL;

      strcpy(tag[nt], "ResamplingRate");
      addr[nt] = &params_.ResamplingRate;
      id[nt++] = INT_SKIP;

      strcpy(tag[nt], "WriteDisplacements");
      addr[nt] = &params_.WriteDisplacements;
      id[nt++] = LOGICAL;

      /*
      strcpy(tag[nt], "StartFromSavedFmax");
      addr[nt] = &params_.StartFromSavedFmax;
      id[nt++] = STRING;
      */

      strcpy(tag[nt], "CatalogInAscii");
      addr[nt] = &params_.CatalogInAscii;
      id[nt++] = LOGICAL;

      strcpy(tag[nt], "DoNotFragment");
      addr[nt] = &params_.DoNotFragment;
      id[nt++] = LOGICAL;

      strcpy(tag[nt], "OutputInH100");
      addr[nt] = &params_.OutputInH100;
      id[nt++] = LOGICAL;

      strcpy(tag[nt], "InputSpectrum_UnitLength_in_cm");
      addr[nt] = &(params_.InputSpectrum_UnitLength_in_cm);
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "FileWithInputSpectrum");
      addr[nt] = params_.FileWithInputSpectrum;
      id[nt++] = STRING;

      strcpy(tag[nt], "WDM_PartMass_in_kev");
      addr[nt] = &(params_.WDM_PartMass_in_kev);
      id[nt++] = DOUBLE;

      strcpy(tag[nt], "TabulatedEoSfile");
      addr[nt] = params_.TabulatedEoSfile;
      id[nt++] = STRING;

      for (j=0; j<nt; j++)     /* All logical tags are FALSE by default */
	if (id[j]==LOGICAL)
	  *((int *) addr[j]) = 0;


      if((fd = fopen(params_.Paramfile, "r")))
	{
	  while(!feof(fd))
	    {
	      *buf = 0;
	      ufo  = fgets(buf, BLENGTH, fd);
	      number_of_fields = sscanf(buf, "%s%s%s%s", buf1, buf2, buf3, buf4);
	      if(number_of_fields < 1)
		continue;

	      if(buf1[0] == '#' || buf1[0]=='%')
		continue;

	      /* searches for a match with the tag list */
	      for(i = 0, j = -1; i < nt; i++)
		if(strcmp(buf1, tag[i]) == 0)
		  {
		    j = i;
		    tag[i][0] = 0;
		    break;
		  }

	      if(j >= 0)
		{
		  switch (id[j])
		    {
		    case DOUBLE:
		      if (number_of_fields<2)
			j=-10;
		      else
			*((double *) addr[j]) = atof(buf2);
		      break;

		    case STRING:
		      if (number_of_fields<2)
			j=-10;
		      else
			strcpy(addr[j], buf2);
		      break;

		    case INT:
		      if (number_of_fields<2)
			j=-10;
		      else
			*((int *) addr[j]) = atoi(buf2);
		      break;

		    case INT_SKIP:
		      if (number_of_fields<2)
			j=-10;
		      else
			*((int *) addr[j]) = atoi(buf2);
		      break;

		    case INT3:
		      if (number_of_fields<4)
			j=-10;
		      else
			{
			  *((int *)addr[j]  ) = atoi(buf2);
			  *((int *)addr[j]+1) = atoi(buf3);
			  *((int *)addr[j]+2) = atoi(buf4);
			}
		      break;

		    case LOGICAL:
		      *((int *) addr[j]) = 1;
		      break;

		    default:
		      break;

		    }

		  if (j<0)
		    {
		      printf("ERROR in file %s:   Tag %s has no argument.\n",
			     params_.Paramfile, buf1);
		      return 1;
		    }
		}
	      else
		{
		  printf("ERROR in file %s:   Tag %s is unknown.\n",
			 params_.Paramfile, buf1);
		  return 1;
		}
	    }
	  fclose(fd);

	}
      else
	{
	  printf("ERROR: Parameter file %s not found.\n", params_.Paramfile);
	  return 2;
	}

      /* Checks that all parameters are found */
      for(i = 0; i < nt; i++)
	{
	  if(*tag[i])
	    {
	      if (id[i]==LOGICAL || id[i]==INT_SKIP) 
		*((int *) addr[i]) = 0;
	      else
		{
		  printf("ERROR: I miss a value for tag '%s' in parameter file '%s'.\n", tag[i], params_.Paramfile);
		  return 1;
		}

	    }
	}


      /* Now opens and reads the list of wanted outputs */
      outputs_.nout=0;
      if((fd = fopen(params_.OutputList, "r")))
	while(!feof(fd))
	  {
	    *buf = 0;
	    ufo  = fgets(buf, BLENGTH, fd);
	    if (buf[0] != '#' && buf[0] != '%' && sscanf(buf,"%lf",&z) != EOF)
	      outputs_.zout[outputs_.nout++]=z;
	  }
      else
	{
	  printf("Error: file %s not found\n",params_.OutputList);
	  return 1;
	}

      if (outputs_.nout<=0)
	{
	  printf("Error in reading output list in file %s\n",params_.OutputList);
	  return 1;
	}

      if (outputs_.nout>MAXOUTPUTS)
	{
	  printf("Error: too many required outputs, increase MAXOUTPUTS\n");
	  return 1;
	}


#ifndef NON_CUBIC
      params_.GridSize[1]=params_.GridSize[2]=params_.GridSize[0];
#endif

      /* lengths of character arrays to ease reading from fortran */
      space_pad(params_.Paramfile);
      space_pad(params_.RunFlag);
      space_pad(params_.OutputList);
      //space_pad(params_.ReadDensityField);
      //space_pad(params_.StartFromSavedFmax);
      space_pad(params_.FileWithInputSpectrum);
      space_pad(params_.TabulatedEoSfile);

    }

  /* processor 0 broadcasts the parameters to all other processors */
  MPI_Bcast(&params_.Omega0, sizeof(params_), MPI_BYTE, 0, MPI_COMM_WORLD);
  MPI_Bcast(&outputs_.zout[0], sizeof(outputs_), MPI_BYTE, 0, MPI_COMM_WORLD);

  return 0;
}


void space_pad(char *string)
{
  int j;

  for (j=0; j<BLENGTH; j++)
    if (string[j]=='\0')
      break;
  
  for (; j<BLENGTH; j++)
    string[j]=' ';

}
