!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

subroutine distribute()

!
! Distributes the Fmax and Vmax fields to all processors
!

  use numbers
  use fmax_grids
  use pinocchio_module
  use fragment_grids
  use fmax_main_memory
  use final_products
  use fragment_matrices
  implicit none

  integer(kind=singleI)::global_z
  integer(kind=singleI),allocatable,dimension(:)::belongs_local,belongs_global

  integer(kind=singleI)::i1,i2,i3,j1,j2,k1,k2,Q,k1start,k2start, &
       main_proc,second_proc,sender,sub_plane_size,local_z,second_z,this_z,quantity,Lgrid_frag12
  character(len=24):: fdate
  integer(kind=singleI)::Lgrid_larg(3)
  integer(kind=singleI),external::find_start,find_length

  real(kind=doubleR),allocatable,dimension(:,:)::sub_plane


  write(info,'(''['',a24,''] Starting to re-distribute planes to processors...'')') fdate()
  call message(trim(info))
  call MPI_BARRIER(MPI_COMM_WORLD,ierr)

  Lgrid_larg=Lgrid/nbox+1+2*safe
  Lgrid_frag12=Lgrid_frag(1)*Lgrid_frag(2)

! here it builds a map of which processor owns which plane

  allocate(belongs_local(0:Lgrid(3)-1))
  allocate(belongs_global(0:Lgrid(3)-1))
  belongs_local=0

  do global_z=0,Lgrid(3)-1

     if (global_z>=Lgrid_z_start .and. global_z<Lgrid_z_start+Lgrid_local(3)) then
        belongs_local(global_z)=proc_id
     endif

  enddo

  call MPI_REDUCE(belongs_local, belongs_global, Lgrid(3), MPI_INTEGER, MPI_SUM, 0, MPI_COMM_WORLD, ierr)
  call MPI_BCAST(belongs_global, Lgrid(3), MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)

  deallocate(belongs_local)

! allocate sub-plane to communicate
! it is allocated with the largest possible size
  total_allocated = total_allocated + 8*Lgrid_larg(1)*Lgrid_larg(2)
  max_allocated   = max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating sub-plane...')
  allocate(sub_plane(0:Lgrid_larg(1)-1,0:Lgrid_larg(2)-1))
  if(alloc_verbose) then
     call message ('sub-plane allocated')
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  sub_plane=0.0
  sub_plane_size=Lgrid_larg(1)*Lgrid_larg(2)

! loop on global planes
  i3=0                    ! counter of sub-boxes, updated at the end of the do cycle
  do global_z=0,Lgrid(3)-1

! loop on all sub-planes
     do i2=0,nbox(2)-1
        k2start=find_start(Lgrid(2),nbox(2),i2)-safe     ! starting (2) coordinate including safety boundary
        do i1=0,nbox(1)-1
           k1start=find_start(Lgrid(1),nbox(1),i1)-safe  ! starting (1) coordinate including safety boundary

! find the main and second processors that will need the sub-plane
           call find_proc(global_z, i1, i2, i3, main_proc, second_proc, local_z, second_z)

! if the plane belongs to this processor then I have to distribute it
           if (belongs_global(global_z)==proc_id) then


! loop on the quantities to distribute
              Q=3
              do quantity=0,Q

! store the sub-plane to communicate  - DIMENSION HERE IS Lgrid_larg TO BE MOSTLY GENERAL
                 do j2=0,Lgrid_larg(2)-1

                    k2=j2+k2start
                    if (k2<0) k2=k2+Lgrid(2)
                    if (k2>=Lgrid(2)) k2=k2-Lgrid(2)

                    do j1=0,Lgrid_larg(1)-1

                       k1=j1+k1start
                       if (k1<0) k1=k1+Lgrid(1)
                       if (k1>=Lgrid(1)) k1=k1-Lgrid(1)

                       select case(quantity)
                       case (0)
                          sub_plane(j1,j2)=Fmax(k1,k2,global_z-Lgrid_z_start)
                       case (1:3)
                          sub_plane(j1,j2)=vmax(quantity,k1,k2,global_z-Lgrid_z_start)
                       end select

                    enddo
                 enddo

! if the main processor is proc_id, then simply store the plane
                 if (main_proc==proc_id) then
                    select case (quantity)
                    case (0)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             Fmax_frag(1+k1+k2*Lgrid_frag(1)+local_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    case (1)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             vx(1+k1+k2*Lgrid_frag(1)+local_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    case (2)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             vy(1+k1+k2*Lgrid_frag(1)+local_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    case (3)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             vz(1+k1+k2*Lgrid_frag(1)+local_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    end select

                 else
! else send it to main_proc
                    call MPI_SEND(sub_plane, sub_plane_size, MPI_DOUBLE_PRECISION, main_proc, 0, MPI_COMM_WORLD, ierr)
                 endif

! if the secondary processor is proc_id, then simply store the plane
                 if (second_proc==proc_id) then
                    select case (quantity)
                    case (0)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             Fmax_frag(1+k1+k2*Lgrid_frag(1)+second_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    case (1)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             vx(1+k1+k2*Lgrid_frag(1)+second_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    case (2)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             vy(1+k1+k2*Lgrid_frag(1)+second_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    case (3)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             vz(1+k1+k2*Lgrid_frag(1)+second_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    end select

                 else if (second_proc>=0) then
! else send it to second_proc
                    call MPI_SEND(sub_plane, sub_plane_size, MPI_DOUBLE_PRECISION, second_proc, 0, MPI_COMM_WORLD, ierr)
                endif

! end of loop on quantity
              enddo

           else
! if the plane does not belong to proc_id, check whether it has to receive it

              Q=3
              do quantity=0,Q

! if the processor is to receive a sub-plane then do it and store it
                 if (main_proc==proc_id .or. second_proc==proc_id) then

                    sender=belongs_global(global_z)
                    call MPI_RECV(sub_plane, sub_plane_size, MPI_DOUBLE_PRECISION, sender, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)
                    if (main_proc==proc_id) then
                       this_z=local_z
                    else
                       this_z=second_z
                    endif

                    select case (quantity)
                    case (0)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             Fmax_frag(1+k1+k2*Lgrid_frag(1)+this_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    case (1)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             vx(1+k1+k2*Lgrid_frag(1)+this_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    case (2)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             vy(1+k1+k2*Lgrid_frag(1)+this_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    case (3)
                       do k2=0,Lgrid_frag(2)-1
                          do k1=0,Lgrid_frag(1)-1
                             vz(1+k1+k2*Lgrid_frag(1)+this_z*Lgrid_frag12)=sub_plane(k1,k2)
                          enddo
                       enddo
                    end select

                 end if

              enddo

           endif

           call MPI_BARRIER(MPI_COMM_WORLD,ierr)

        enddo
     enddo

! updates the z-sub-box index
     if (global_z==find_start(Lgrid(3),nbox(3),i3+1)-1) then
        i3=i3+1
     endif

  enddo

  deallocate(belongs_global)
  deallocate(sub_plane)
  total_allocated = total_allocated - 8*Lgrid_larg(1)*Lgrid_larg(2)
  if(alloc_verbose) then
     call message('de-allocated sub_plane')
  end if

  return
end subroutine distribute


subroutine find_proc(global_z, i1, i2, i3, main_proc, second_proc, local_z, second_z)

! given the z-plane, find the main and secondary processor
! to which the sub-plane must be communicated

  use pinocchio_module
  use fmax_grids
  use fragment_grids
  implicit none

  integer(kind=singleI),intent(in)::global_z, i1, i2, i3
  integer(kind=singleI),intent(out)::main_proc, second_proc, local_z, second_z
  integer(kind=singleI)::TL
  integer(kind=singleI),external::find_start,find_length

! sub-planes are distributed to processors first in z, then in y and x 
  main_proc=i3+i2*nbox(3)+i1*nbox(3)*nbox(2)

! this is the local z variable for the processor
  local_z=global_z-find_start(Lgrid(3),nbox(3),i3)+safe
  TL=find_length(Lgrid(3),nbox(3),i3)

! the second processor to communicate the sub_plane to is decided
! using local_z
  if (local_z<2*safe) then            ! lower border, second processor is the previous one in z

     if (i3>0) then
        second_proc=main_proc-1
        second_z=find_length(Lgrid(3),nbox(3),i3-1)+local_z
     else
        second_proc=main_proc+nbox(3)-1
        second_z=find_length(Lgrid(3),nbox(3),nbox(3)-1)+local_z
     endif

  else if (local_z>=TL) then   ! higher border, second processor is the next one in z

     if (i3<nbox(3)-1) then
        second_proc=main_proc+1
        second_z=local_z-TL
     else
        second_proc=main_proc-nbox(3)+1
        second_z=local_z-TL
     endif

  else
     second_proc=-1
     second_z=-1
  endif

  return
end subroutine find_proc


subroutine find_subbox(ibox)

! find the sub-box of processor proc_id

  use pinocchio_module
  implicit none
  integer(kind=singleI),dimension(3),intent(out)::ibox
  integer(kind=singleI)::kk,nn

  if (proc_id >= product(nbox)) then
     ibox=-1
     return
  endif

  nn=nbox(2)*nbox(3)
  ibox(1)=proc_id/nn
  kk=proc_id-ibox(1)*nn
  ibox(2)=kk/nbox(3)
  ibox(3)=kk-ibox(2)*nbox(3)

  return
end subroutine find_subbox


function find_start(L,n,ibox)

! finds the starting coordinate of a subbox, given the grid length L, the number of subboxes n and the subbox id ibox

  use pinocchio_module
  implicit none
  integer(kind=singleI)::find_start
  integer(kind=singleI),intent(in)::L,n,ibox
  integer(kind=singleI)::LL,MM

  if (n==1) then
     find_start=0
  else

     LL=L/n
     MM=mod(L,n)
     if (ibox==0) then
        find_start=0
     else if (ibox<=MM) then
        find_start=ibox*(LL+1)
     else
        find_start=ibox*LL+MM
     endif

  endif

  return
end function find_start

function find_length(L,n,ibox)

! finds the length of a subbox, given the grid length L, the number of subboxes n and the subbox id ibox

  use pinocchio_module
  implicit none
  integer(kind=singleI)::find_length
  integer(kind=singleI),intent(in)::L,n,ibox
  integer(kind=singleI)::LL,MM

  if (n==1) then 
     find_length=L
  else

     LL=L/n
     MM=mod(L,n)
     if (ibox<MM) then
        find_length=LL+1
     else
        find_length=LL
     endif

  endif

  return
end function find_length
