!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


!!!#define OLD_LOOP

#define SIMPLE_LOOP

!!!#ifdef USE_GENIC
!!!#define NORMAL_ORDER
!!!#else
!!!#define SIMPLE_LOOP
!!!#endif

! This file contains the wrappers for fftw

subroutine compute_derivative(first_derivative, second_derivative)


! This code computes a derivative of the (peculiar rescaled) gravitational potential
! using ffts.
! The k-space potential is assumed to be in vector, 
! first_derivative and second_derivative are integers used in the Green function
! they can assume the following values:
!
!   0 = no derivative
!   1 = differentiate over x-direction
!   2 = differentiate over y-direction
!   3 = differentiate over z-direction

  use pinocchio_module
  use fmax_fftw_variables

  implicit none
  character(len=24):: fdate
  external fdate

  integer(kind=singleI),intent(in)::first_derivative, second_derivative

! local variables
  logical::swap
  integer(kind=singleI),parameter::Ne=16
  integer(kind=singleI)::local_x,local_y,local_z,ix,iy,iz,ixx, iyy, izz
  integer(kind=singleI)::index,nxhalf,nyhalf,nzhalf
  integer(kind=singleI),external :: init_random
  integer(kind=doubleI),external :: index_kspace
#ifdef OLD_LOOP
  integer(kind=singleI)::iloop,istart
#endif
!
  real(kind=doubleR):: kx,ky,kz,kxnorm,kynorm,kznorm,green &
       ,k_squared,k_module,smoothing,tmp, norm
  real(kind=doubleR),dimension(0:3)::diff_comp=1.d0
  real(kind=doubleR),external::power,random,greens_function
!

! normalization of the k-space modes
  norm = 2.d0*pi**2 / cell_size**3 * dble(nx_fftw)*dble(ny_fftw)*dble(nz_fftw)

! k vectors
  kxnorm   = 2.*pi/dble(nx_fftw)
  kynorm   = 2.*pi/dble(ny_fftw)
  kznorm   = 2.*pi/dble(nz_fftw)

! Nyquist frequencies
  nxhalf = nx_fftw/2
  nyhalf = ny_fftw/2
  nzhalf = nz_fftw/2

! for first derivatives the real and imaginary parts must be swapped
  swap=.false.
  if ((first_derivative==0 .and. second_derivative> 0) .or. &
      (first_derivative> 0 .and. second_derivative==0))     &
       swap=.true.


! NB: ix, iy and iz are global coordinates of the box (from 0 to N-1)
!     ix: [0,N/2]
!     iy, iz: [0,N-1]
!     iyy and izz are unfolded, they run from -N/2+1 to N/2 (ixx = ix)
!     local_x, local_y and local_z are local coordinates of the slice
!     (for the x and z coordinates they are the same as ix and iz)


! loop over k-space indices

#ifdef OLD_LOOP

! This loop was used in pinocchio-2.2; it is inefficient and there is no need to use it

! loop over planes in y
!  write(info,'(''['',a24,''] real_derivative: fixing k-space quantity '')') fdate()
!  call message (trim(info))
  loopy: do local_y=0,local_ny_after_transform-1
     iy  = local_y + local_y_start_after_transform
     iyy = iy ; if (iy > nyhalf) iyy = iy-ny_fftw
     ky  = kynorm*dble(iyy)

! in this case the loop is done on concentric frames in the ixx-izz space
     loop: do iloop = 0, max(nzhalf,nxhalf)
        
        loopz: do izz = -iloop,iloop

           if(abs(izz) .eq. iloop) then
              istart = 0
           else
              istart = iloop
           endif
           
           loopx: do ixx = istart, iloop
              
! the loop cycles if the point is not on the iloop frame (which means that 
! either ixx or abs(izz) are not equal to iloop) or if the loop is going on 
! non-existent points (this happens for non-cubic boxes)
              if (abs(izz) /= iloop .and. ixx /= iloop) then
                 write (*,*) ' error in loop ',iloop,ixx,izz
                 call abort_code(' error ')
              endif
              
              if ( izz <= -nzhalf .or. izz >nzhalf  &
                   .or. ixx < 0 .or. ixx > nxhalf) cycle 
              
              iz = izz ; if (izz<0) iz = izz + nz_fftw
              local_z = iz
              kz = kznorm*dble(izz)
              
              ix = ixx
              local_x = ix
              kx = kxnorm*dble(ixx)
              
#endif


#ifdef SIMPLE_LOOP

! This loop is correct for transposed order in FFTW

  do local_z = 0, nz_fftw-1
     iz  = local_z
     izz = iz ; if (iz > nzhalf) izz = iz - nz_fftw
     kz  = kznorm*dble(izz)

     do local_y = 0, local_ny_after_transform-1
        iy  = local_y + local_y_start_after_transform
        iyy = iy ; if (iy > nyhalf) iyy = iy - ny_fftw
        ky  = kynorm*dble(iyy)

        do local_x = 0, nxhalf
           ix  = local_x
           ixx = ix
           kx  = kxnorm*dble(ixx)

#endif


#ifdef NORMAL_ORDER

! This loop is correct in case normal order in FFTW is asked
! use it to achieve full consistency with GenIC runs, but it is less efficient
! do not use it without GenIC

! NB: here we have to treat x as z and vice-versa, though it is boring

  do local_x = 0, local_nz-1
     ix  = local_x + local_z_start
     ixx = ix ; if (ix > nzhalf) ixx = ix - nz_fftw
     kx  = kznorm*dble(ixx)

     do local_y = 0, ny_fftw-1
        iy  = local_y
        iyy = iy ; if (iy > nyhalf) iyy = iy - ny_fftw
        ky  = kynorm*dble(iyy)

        do local_z = 0, nxhalf
           iz  = local_z
           izz = iz
           kz  = kxnorm*dble(izz)

#endif

              k_squared  = kx**2 + ky**2 + kz**2
              k_module   = sqrt(k_squared)

! corresponding index of real part in vector (imaginary in index + 1)
              index = index_kspace(local_x, local_y, local_z)

              if (k_squared /= 0.) then

! Gaussian smoothing window
                 smoothing = exp(-0.5d0*k_squared*Rsm_grid**2)

!                 vector(index)   = vector(index)   * smoothing
!                 vector(index+1) = vector(index+1) * smoothing

! the components are stored in the vectors that control the differentiation
                 diff_comp = (/1.d0,kx,ky,kz/)

                 green     = greens_function(diff_comp, k_squared, first_derivative, second_derivative)

                 vector(index)   =  green * smoothing * vector(index)    ! real
                 vector(index+1) =  green * smoothing * vector(index+1)  ! imaginary
              else
! fundamental mode
                 vector(index)   = 0.
                 vector(index+1) = 0.
              endif
              
              if (swap) then
                 tmp             = vector(index+1)
                 vector(index+1) = vector(index)     ! imaginary
                 vector(index)   = - tmp             ! real
              endif

!              if (first_derivative==2) then 
!                 write(50,'(3(1x,i3),1x,i9,9(1x,g15.6))')ix,iy,iz,(index-1)/2,vector(index),vector(index+1),k_module
!              endif

#ifdef OLD_LOOP

           enddo loopx
        enddo loopz
     enddo loop
     
  enddo loopy

#endif

#if defined(SIMPLE_LOOP) || defined(NORMAL_ORDER)

        enddo
     enddo
  enddo

#endif


! reverse complex_to_real FFT
  write(info,'(''['',a24,''] compute_derivative: starting fft '')') fdate()
  call message (trim(info))

  call reverse_transform ()

!  write(40+proc_id,*)vector
  
  write(info,'(''['',a24,''] compute_derivative: done fft '')') fdate()
  call message (trim(info))

  return
end subroutine compute_derivative


function index_rspace(local_x, local_y, local_z)

! index of vector in real-space

  use numbers
  use fmax_fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  implicit none
  integer(kind=singleI), intent(in) :: local_x, local_y, local_z
  integer(kind=doubleI) :: index_rspace

#ifdef NORMAL_ORDER
  index_rspace = 1 + local_z + local_y * (nx_fftw+2) + local_x * (nx_fftw+2) * ny_fftw  
#else
  index_rspace = 1 + local_x + local_y * (nx_fftw+2) + local_z * (nx_fftw+2) * ny_fftw
#endif

end function index_rspace


function index_kspace(local_x, local_y, local_z)

! index of vector in k-space

  use numbers
  use fmax_fftw_variables, ONLY: nx_fftw, ny_fftw, nz_fftw
  implicit none
  integer(kind=singleI), intent(in) :: local_x, local_y, local_z
  integer(kind=doubleI) :: index_kspace

#ifdef NORMAL_ORDER
  index_kspace = 1 + 2*local_z + (nx_fftw+2)*(local_y + local_x*ny_fftw)
#else
  index_kspace = 1 + 2*local_x + (nx_fftw+2)*(local_z + local_y*nz_fftw)
#endif

end function index_kspace

 

function greens_function(diff_comp, k_squared, first_derivative, second_derivative)

! green function for the differentiation

  use numbers
  implicit none
  real(kind=doubleR) :: greens_function
  integer(kind=singleI),intent(in)::first_derivative,second_derivative
  real(kind=doubleR),intent(in)::diff_comp(0:3)
  real(kind=doubleR),intent(in)::k_squared


! in this case the greens_function is simply 1
  if (first_derivative == -1 .and. second_derivative == -1) then
     greens_function = 1.d0
     return
  end if

  if (first_derivative==0 .and. second_derivative==0) then
     greens_function = - diff_comp(first_derivative)*diff_comp(second_derivative)/k_squared
  else
     greens_function =   diff_comp(first_derivative)*diff_comp(second_derivative)/k_squared
  endif

  return
end function greens_function


subroutine init_fftw(igrid)

! Creates the plans for FFTW and gets the number of planes per processor

  use fmax_grids
  use pinocchio_module
  use fmax_fftw_variables
 
  implicit none
  integer(kind=singleI),intent(in)::igrid

! else create plans
  if (igrid==1) then
     nx_fftw = Lgrid(1) ; ny_fftw = Lgrid(2) ; nz_fftw = Lgrid(3)

     call create_plan()

! local grid in real space for each processor
     saved_forward_plan   = forward_plan
     saved_reverse_plan   = reverse_plan
     Lgrid_local(1:2)     = Lgrid(1:2)
     Lgrid_local(3)       = local_nz
     Lgrid_z_start        = local_z_start
     npart_local          = Lgrid_local(1)*Lgrid_local(2)*Lgrid_local(3)
! local grid in Fourier space for each processor
     Lgrid_fft(1)         = Lgrid(1)/2+1
     Lgrid_fft(2)         = local_ny_after_transform
     Lgrid_fft(3)         = Lgrid(3)
     Lgrid_y_start        = local_y_start_after_transform
     npart_fft            = total_local_size

  else

     call abort_code("MULTIPLE GRIDS NOT IMPLEMENTED YET")

  endif

  return
end subroutine init_fftw


subroutine create_plan()

! Creates a plan for fftw

  use pinocchio_module
  use fmax_fftw_variables
  implicit none

  call RFFTW3D_F77_MPI_CREATE_PLAN(forward_plan ,MPI_COMM_WORLD  & 
       , nx_fftw, ny_fftw, nz_fftw &
       , FFTW_REAL_TO_COMPLEX , FFTW_ESTIMATE + FFTW_IN_PLACE)

  call RFFTW3D_F77_MPI_CREATE_PLAN(reverse_plan ,MPI_COMM_WORLD &
       , nx_fftw, ny_fftw, nz_fftw &
       , FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE)

  call RFFTWND_F77_MPI_LOCAL_SIZES(forward_plan, local_nz, local_z_start, &
       local_ny_after_transform, local_y_start_after_transform, total_local_size)


!!!  call rfftw3d_f77_create_plan(forward_plan & !,MPI_COMM_WORLD  & 
!!!       , nx_fftw, ny_fftw, nz_fftw &
!!!       , FFTW_REAL_TO_COMPLEX , FFTW_ESTIMATE + FFTW_IN_PLACE)
!!!
!!!  call rfftw3d_f77_create_plan(reverse_plan & !,MPI_COMM_WORLD &
!!!       , nx_fftw, ny_fftw, nz_fftw &
!!!       , FFTW_COMPLEX_TO_REAL, FFTW_ESTIMATE + FFTW_IN_PLACE)
!!!
!!!  local_nz = nz_fftw
!!!  local_z_start = 0
!!!  local_ny_after_transform = ny_fftw
!!!  local_y_start_after_transform = 0
!!!  total_local_size = (nx_fftw +2) * ny_fftw * nz_fftw


  return
end subroutine create_plan


subroutine forward_transform()

! wrapper for a forward transform with fftw

  use pinocchio_module
  use fmax_fftw_variables
  implicit none

! timing
  cputmp=MPI_WTIME()

  call rfftwnd_f77_mpi(forward_plan, 1, vector, fftw_work, use_work_fftw,  &
#ifdef NORMAL_ORDER
       FFTW_NORMAL_ORDER)
#else
       FFTW_TRANSPOSED_ORDER)
#endif

! timing
  cputime(1)=cputime(1)+MPI_WTIME()-cputmp

  return
end subroutine forward_transform


subroutine reverse_transform ()

! wrapper for a reverse transform with fftw

  use pinocchio_module
  use fmax_fftw_variables
  implicit none
  real(kind=doubleR) :: norm

! timing
  cputmp=MPI_WTIME()

! normalization
  norm = 1.d0/dble(nx_fftw)/dble(ny_fftw)/dble(nz_fftw)

  call rfftwnd_f77_mpi(reverse_plan, 1, vector, fftw_work, use_work_fftw, &
#ifdef NORMAL_ORDER
       FFTW_NORMAL_ORDER)
#else
       FFTW_TRANSPOSED_ORDER)
#endif

  vector = vector * norm

! timing
  cputime(1)=cputime(1)+MPI_WTIME()-cputmp

  return
end subroutine reverse_transform

#ifdef DO_NOT_USE_GENIC

! This is the routine for the generation of initial conditions used by pinocchio-2.2

subroutine generate_one_density()

! reads or generates a potential in k-space

  use pinocchio_module
  use fmax_fftw_variables
  use fmax_main_memory

  implicit none
  character(len=24):: fdate
  external fdate

! local variables
  integer(kind=singleI),parameter::Ne=16
  integer(kind=singleI)::local_x,local_y,local_z,ix,iy,iz,ixx, iyy, izz,iyp,istart
  integer(kind=singleI)::index,nxhalf,nyhalf,nzhalf
  integer(kind=singleI),external :: init_random
  integer(kind=singleI)::iloop
  integer(kind=doubleI),external :: index_kspace
!
  real(kind=doubleR):: kx,ky,kz,kxnorm,kynorm,kznorm,spectrum &
       ,amplitude,k_squared,k_module,phase,random1, random2, k_physical, norm, pow, smo
  real(kind=doubleR),external::power,random,greens_function
!

! ids of the y-planes of the slice
  integer(kind=singleI), allocatable:: plane_id(:)
  integer(kind=singleI):: streamnum
  integer(kind=singleI)::stream

! normalization of the k-space modes
  norm = 2.d0*pi**2 / cell_size**3 * dble(nx_fftw)*dble(ny_fftw)*dble(nz_fftw)

! k vectors
  kxnorm   = 2.*pi/dble(nx_fftw)
  kynorm   = 2.*pi/dble(ny_fftw)
  kznorm   = 2.*pi/dble(nz_fftw)

! Nyquist frequencies
  nxhalf = nx_fftw/2
  nyhalf = ny_fftw/2
  nzhalf = nz_fftw/2

! initialise plane-id's for determining random stream so that large scale power 
! is independent of resolution
 index = 0
  allocate(plane_id(ny_fftw))
  plane_id(1) = index
  do iy=1, ny_fftw/2
     iyp = -iy + ny_fftw
     index = index + 1 ;  plane_id(iy+1)  = index
     if(iy .lt. ny_fftw/2) then
        index = index + 1 ;  plane_id(iyp+1) = index
     endif
  enddo


! NB: ix, iy and iz are global coordinates of the box (from 0 to N-1)
!     ix: [0,N/2]
!     iy, iz: [0,N-1]
!     iyy and izz are unfolded, they run from -N/2+1 to N/2 (ixx = ix)
!     local_x, local_y and local_z are local coordinates of the slice
!     (for the x and z coordinates they are the same as ix and iz)


! loop over k-space indices

! loop over planes in y
  write(info,'(''['',a24,''] generate_density: init random numbers '')') fdate()
  call message (trim(info))

  loopy: do local_y=0,local_ny_after_transform-1
     iy  = local_y + local_y_start_after_transform
     iyy = iy ; if (iy > nyhalf) iyy = iy-ny_fftw
     ky  = kynorm*dble(iyy)

! start random sequence for this y-plane
     streamnum  = plane_id(iy + 1)
     stream     = init_random(saved_seed,ThisGrid,streamnum)
     
! in this case the loop is done on concentric frames in the ixx-izz space
     loop: do iloop = 0, max(nzhalf,nxhalf)
        
        loopz: do izz = -iloop,iloop

           if(abs(izz) .eq. iloop) then
              istart = 0
           else
              istart = iloop
           endif
           
           loopx: do ixx = istart, iloop
              
! the loop cycles if the point is not on the iloop frame (which means that 
! either ixx or abs(izz) are not equal to iloop) or if the loop is going on 
! non-existent points (this happens for non-cubic boxes)
              if (abs(izz) /= iloop .and. ixx /= iloop) then
                 write (*,*) ' error in loop ',iloop,ixx,izz
                 call abort_code(' error ')
              endif
              
              if ( izz <= -nzhalf .or. izz >nzhalf  &
                   .or. ixx < 0 .or. ixx > nxhalf) cycle 
              
              iz = izz ; if (izz<0) iz = izz + nz_fftw
              local_z = iz
              kz = kznorm*dble(izz)
              
              ix = ixx
              local_x = ix
              kx = kxnorm*dble(ixx)
              
              k_squared  = kx**2 + ky**2 + kz**2
              k_module   = sqrt(k_squared)
              k_physical = k_module/cell_size
              
! corresponding index of real part in vector (imaginary in index + 1)
              index = index_kspace(local_x, local_y, local_z)

              random1   = random(stream)
              random2   = random(stream)

              if (k_squared /= 0.) then

! power spectrum with high- and low-frequency cutoffs
                 pow=power(k_physical)
                 smo=exp(-(k_module/upper_k_cutoff)**Ne)

                 if (k_module<upper_k_cutoff*1.2) then
                    spectrum = pow * norm * smo
                    amplitude = sqrt(-spectrum*log(random1))
                 else
                    spectrum = 0.0d0
                    amplitude = 0.0d0
                 endif

!                 if (lower_k_cutoff > 0.) spectrum = spectrum &
!                      *(1.d0 - exp(-(k_module/lower_k_cutoff)**Ne))

! phase is set to 0 for k-vectors on Nyquist edge
                 if ((ix==0 .or. ix==nxhalf) .and. &
                      (iy==0 .or. iy==nyhalf) .and. &
                      (iz==0 .or. iz==nzhalf)) then
                    phase = 0.d0
                 else
                    phase = 2.d0*pi*random2
                 endif

                 vector(index)   = amplitude*cos(phase)
                 vector(index+1) = amplitude*sin(phase)

              endif


!              write(50+proc_id,'(99(1x,e15.8))')k_module,pow,smo,amplitude,phase,vector(index),vector(index+1)


           enddo loopx
        enddo loopz
     enddo loop

  enddo loopy


  deallocate(plane_id)

  write(info,'(''['',a24,''] generate_one_density: done'')') fdate()
  call message (trim(info))

  return
end subroutine generate_one_density

#else

! This calls the generation of ICs extrapolated from 2LPTic by Roman Scoccimarro,
! that part is equal to NGenIC by Volker Springel

subroutine generate_one_density()

  use pinocchio_module
  use fmax_fftw_variables
  use fmax_grids
  implicit none

#ifndef UNDERSCORE
  integer(kind=singleI),external::GenIC
#else
  integer(kind=singleI),external::GenIC_
#endif
  real(kind=doubleR),target,allocatable,dimension(:)::vec_genic
  real(kind=doubleR),pointer::pointer_to_vector
#ifndef UNDERSCORE
  common /comm_genic/pointer_to_vector
#else
  common /comm_genic_/pointer_to_vector
#endif

  character(len=24),external :: fdate
  integer(kind=singleI)::j

  total_allocated = total_allocated + sizeof_vector*npart_fft
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating vec_genic...')
  allocate(vec_genic(0:npart_fft-1))
  if (alloc_verbose) then
     call message ('vec_genic allocated')
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif
  pointer_to_vector=>vec_genic(0)

#ifndef UNDERSCORE
  j=GenIC(npart_fft, local_ny_after_transform, local_y_start_after_transform, proc_id, nproc, 0)
#else
  j=GenIC_(npart_fft, local_ny_after_transform, local_y_start_after_transform, proc_id, nproc, 0)
#endif

  vector=vec_genic
  nullify(pointer_to_vector)
  deallocate(vec_genic)
  total_allocated = total_allocated - sizeof_vector*npart_fft
  if (alloc_verbose) call message('de-allocated vec_genic')


  write(info,'(''['',a24,''] generate_one_density_with_GenIC: done'')') fdate()
  call message (trim(info))

  return
end subroutine generate_one_density
#endif

