!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

! Modules for fmax

module fmax_grids

  use numbers

! number of grids
  integer(kind=singleI)::ngrid
! 3D lengths of the slice assigned to the processor
  integer(kind=singleI),dimension(3)            :: Lgrid_local
! 3D length of the k-space slice assigned to the processor
  integer(kind=singleI),dimension(3)            :: Lgrid_fft
! Floating-point value of Lgrid_overlap (can be <1)
! starting z-plane for the real-space slice
  integer(kind=singleI)                         :: Lgrid_z_start
! starting y-plane for the k-space slice
  integer(kind=singleI)                         :: Lgrid_y_start
! comoving box sizes
  real(kind=doubleR),dimension(3)               :: BoxSizeGrids
! comoving lower cutoff to apply to the box
  real(kind=doubleR)                            :: k_cutoff
! number of particles in the real-space slice
  integer(kind=doubleI)                         :: npart_local
! number of particles in the k-space slice
  integer(kind=doubleI)                         :: npart_fft
! fftw plans
  integer(kind=doubleI) :: saved_forward_plan, saved_reverse_plan
! Files with the input initial conditions (if relevant)
!  character(len=30),allocatable,dimension(:)      :: InputDensityFile
end module fmax_grids


module fmax_fftw_variables
! These module contains the variables that are passed to the fftw wrapper.
! They are re-initializes each time that a grid is considered
  use numbers
!  include '/opt/local/fftw/64/current/include/fftw_f77.i'
  include 'fftw_f77.i'
! grid size (saved in Lgrid(igrid,:))
  integer(kind=singleI)     :: nx_fftw, ny_fftw, nz_fftw
! fftw plans (saved in saved_forward_plan and saved_reverse_plan)
  integer(kind=doubleI)     :: forward_plan, reverse_plan
! number of real-space z-planes and starting z-plane for the slice
!   (saved in Lgrid_local(igrid,3) and Lgrid_z_start) 
  integer(kind=singleI)     :: local_nz, local_z_start 
! number of k-space y-planes and starting y-plane for the slice
!   (saved in Lgrid_fft(igrid,2) and Lgrid_y_start)
  integer(kind=singleI)     :: local_ny_after_transform, local_y_start_after_transform 
! total number of particles in the slice (saved in npart_fft)
  integer(kind=singleI)     :: total_local_size
! flag for extra fftw-related output
  integer(kind=singleI)     :: fftw_verbose
! grid considered (equivalent to igrid)
  integer(kind=singleI)     :: ThisGrid

! These are the largest matrices to be allocated by fmax
! vector contains the field to process with fftw (either in real- or k-space)
! fftw_work is a workspace for fftw
  real(kind=doubleR),allocatable,dimension(:)::vector
  real(kind=doubleR),allocatable,dimension(:)::fftw_work
  integer(kind=singleI)::sizeof_vector=8,sizeof_fftw_work=8
! this flag tells fft to use the allocated workspace
  integer(kind=singleI),parameter :: use_work_fftw=1

! other stuff
  real(kind=doubleR)::Rsm_grid  ! smoothing radius in grid coordinates
  real(kind=doubleR)::upper_k_cutoff, lower_k_cutoff   ! cutoffs in k-space, grid coordinates
  real(kind=doubleR)::cell_size ! inter-particle distance

end module fmax_fftw_variables


module fmax_smoothing_radii
! smoothing radii used by fmax, theorical and actual variances of the smoothed
! linear density field
  use numbers
  integer(kind=singleI)::nsmooth
  real(kind=doubleR),allocatable,dimension(:)::Smoothing_Radii, Smoothing_Variances, True_variances
end module fmax_smoothing_radii


module final_products
! Fmax and vmax fields used by fragment
  use numbers

  real(kind=doubleR),allocatable,dimension(:,:,:)::Fmax
  real(kind=doubleR),allocatable,dimension(:,:,:,:)::vmax
  integer(kind=singleI)::sizeof_Fmax=8,sizeof_vmax=8*3

end module final_products


module fmax_main_memory
! main matrices to be allocated
  use numbers
! high-res box
  real(kind=doubleR),allocatable,dimension(:)::kdensity
  real(kind=doubleR),allocatable,dimension(:,:,:)::density
  real(kind=doubleR),allocatable,dimension(:,:,:,:)::first_derivatives,second_derivatives
! low-res box
  real(kind=doubleR),allocatable,dimension(:)::kdensity_l
  real(kind=doubleR),allocatable,dimension(:,:,:)::density_l
  real(kind=doubleR),allocatable,dimension(:,:,:,:)::first_derivatives_l,second_derivatives_l

  real(kind=doubleR),allocatable,dimension(:,:,:,:)::VEL_for_displacements

  integer(kind=singleI)::sizeof_kdensity=8,sizeof_density=8, &
       sizeof_firstd=8*3,sizeof_secondd=8*6,sizeof_Rmax=4

  integer(kind=singleI),allocatable,dimension(:,:,:)::Rmax

end module fmax_main_memory

