!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

subroutine write_fmax()

  use pinocchio_module
  use fmax_grids
  use fmax_main_memory
  use final_products
  implicit none

! local variables
  integer(kind=singleI) :: iproc, local_z, global_z
  character(len=80)     :: filename, position
  real(kind=singleR),allocatable :: plane(:,:)

  if (DoNotFragment) allocate(plane(Lgrid(1),Lgrid(2)))

  write(filename,"('Data/Fmax.',a,'.d')")trim(run)

  do iproc=0,nproc-1
     if(iproc .eq. proc_id) then

        if (iproc==0) then
           position="asis"
        else
           position="append"
        endif

#ifdef ASCII_WRITE
        open(10,file=filename,status="unknown",position=trim(position))
#else
        open(10,file=filename,status="unknown",position=trim(position),form='unformatted')
#endif

        if (proc_id==0) then
#ifdef ASCII_WRITE
           write(10,*)Lgrid(1),Lgrid(2),Lgrid(3)
#else
           write(10)Lgrid(1),Lgrid(2),Lgrid(3)
#endif
        endif

! write data plane by plane
        do local_z=0,Lgrid_local(3)-1
           global_z = local_z + Lgrid_z_start

#ifdef ASCII_WRITE
           write(10,*) global_z
           write(10,'(8(1x,e13.6))') Fmax(:,:,local_z)
#else
           write(10) global_z
           if (DoNotFragment) then
              plane=Fmax(:,:,local_z)
              write(10) plane
           else
              write(10) Fmax(:,:,local_z)
           endif
#endif
        enddo

        close(10)

     endif

! MPI barrier
     call MPI_BARRIER(MPI_COMM_WORLD,ierr)
  enddo

  if (DoNotFragment) deallocate(plane)

! exit
  info = ' written file '//trim(filename)
  call message(trim(info))
  return
end subroutine write_fmax


subroutine write_vmax()

  use pinocchio_module
  use fmax_grids
  use fmax_main_memory
  use final_products
  implicit none

! local variables
  integer(kind=singleI)    :: iproc, local_z, global_z,ia
  character(len=80)        :: filename, position
  real(kind=singleR),allocatable :: plane(:,:)

  if (DoNotFragment) allocate(plane(Lgrid(1),Lgrid(2)))

  do ia=1,3

     if (ia==1) then
        write(filename,"('Data/Velx.',a,'.d')")trim(run)
     else if (ia==2) then
        write(filename,"('Data/Vely.',a,'.d')")trim(run)
     else if (ia==3) then
        write(filename,"('Data/Velz.',a,'.d')")trim(run)
     endif

     do iproc=0,nproc-1
        if(iproc .eq. proc_id) then

           if (iproc==0) then
              position="asis"
           else
              position="append"
           endif

#ifdef ASCII_WRITE
           open(10,file=filename,status="unknown",position=trim(position))
#else
           open(10,file=filename,status="unknown",position=trim(position),form='unformatted')
#endif

           if (proc_id==0) then
#ifdef ASCII_WRITE
              write(10,*)Lgrid(1),Lgrid(2),Lgrid(3)
#else
              write(10)Lgrid(1),Lgrid(2),Lgrid(3)
#endif
           endif

! write data plane by plane
           do local_z=0,Lgrid_local(3)-1
              global_z = local_z + Lgrid_z_start

#ifdef ASCII_WRITE
              write(10,*) global_z
              write(10,'(8(1x,e13.6))') vmax(ia,:,:,local_z)
#else
              write(10) global_z
              if (DoNotFragment) then
                 plane=vmax(ia,:,:,local_z)
                 write(10) plane
              else
                 write(10) vmax(ia,:,:,local_z)
              endif
#endif
           enddo

           close(10)

        endif

! MPI barrier
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
     enddo

     info = ' written file '//trim(filename)
     call message(trim(info))

  enddo

  if (DoNotFragment) deallocate(plane)

! exit
  return
end subroutine write_vmax


subroutine write_rmax()

  use pinocchio_module
  use fmax_grids
  use fmax_main_memory
  implicit none

! local variables
  integer(kind=singleI)    :: iproc, local_z, global_z
  character(len=80)        :: filename, position

  write(filename,"('Data/Rmax.',a,'.d')")trim(run)

  do iproc=0,nproc-1
     if(iproc .eq. proc_id) then

        if (iproc==0) then
           position="asis"
        else
           position="append"
        endif

#ifdef ASCII_WRITE
        open(10,file=filename,status="unknown",position=trim(position))
#else
        open(10,file=filename,status="unknown",position=trim(position),form='unformatted')
#endif

! write data plane by plane
        if (proc_id==0) then
#ifdef ASCII_WRITE
           write(10,*)Lgrid(1),Lgrid(2),Lgrid(3)
#else
           write(10)Lgrid(1),Lgrid(2),Lgrid(3)
#endif
        endif

        do local_z=0,Lgrid_local(3)-1
           global_z = local_z + Lgrid_z_start

#ifdef ASCII_WRITE
           write(10,*)global_z
           write(10,'(8(1x,i5))') Rmax(:,:,local_z)
#else
           write(10) global_z
           write(10) Rmax(:,:,local_z)
#endif
        enddo

        close(10)

     endif

! MPI barrier
     call MPI_BARRIER(MPI_COMM_WORLD,ierr)
  enddo

! exit
  info = ' written '//trim(filename)
  call message(trim(info))
  return
end subroutine write_rmax


subroutine write_density()

  use pinocchio_module
  use fmax_grids
  use fmax_main_memory
  use fmax_main_memory
  implicit none

! local variables
  integer(kind=singleI)    :: iproc, local_z, global_z, igrid,ix,iy,L1,L2
  character(len=80)        :: filename, position
  real(kind=doubleR),allocatable::plane(:,:)
#ifdef ASCII_WRITE
  integer(kind=singleI):: i, j
#endif

  igrid=1
!  do igrid=1,ngrid

  if (ResamplingRate==0) then

     write(filename,"('Data/density.g',i1,'.',a,'.d')")igrid,trim(run)

     do iproc=0,nproc-1
        if(iproc .eq. proc_id) then

           if (iproc==0) then
              position="asis"
           else
              position="append"
           endif

#ifdef ASCII_WRITE
           open(10,file=filename,status="unknown",position=trim(position))
#else
           open(10,file=filename,status="unknown",position=trim(position),form="unformatted")
#endif

           if (proc_id==0) then
#ifdef ASCII_WRITE
              write(10,*)Lgrid(1),Lgrid(2),Lgrid(3)
#else
              write(10)Lgrid(1),Lgrid(2),Lgrid(3)
#endif
           endif
! write data plane by plane
           do local_z=0,Lgrid_local(3)-1
              global_z = local_z + Lgrid_z_start

#ifdef ASCII_WRITE
!              do j=0,Lgrid_local(2)-1
!                 do i=0,Lgrid_local(1)-1
!                    write(10,*)i,j,global_z,density(i,j,local_z)
!                 enddo
!              enddo
              write(10,*) global_z
              write(10,'(8(1x,e13.6))') density(:,:,local_z)
#else
              write(10) global_z
              write(10) density(:,:,local_z)
#endif
           enddo

           close(10)

        endif

! MPI barrier
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
     enddo

! exit
     info = ' written '//trim(filename)
     call message(trim(info))

!  enddo
  else

     L1=Lgrid(1)/ResamplingRate
     L2=Lgrid(2)/ResamplingRate
     allocate(plane(0:L1-1,0:L2-1))

     write(filename,"('Data/density.g',i1,'.',a,'.toberesampled.d')")igrid,trim(run)

     do iproc=0,nproc-1
        if(iproc .eq. proc_id) then

           if (iproc==0) then
              position="asis"
           else
              position="append"
           endif

#ifdef ASCII_WRITE
           open(10,file=filename,status="unknown",position=trim(position))
#else
           open(10,file=filename,status="unknown",position=trim(position),form="unformatted")
#endif

           if (proc_id==0) then
#ifdef ASCII_WRITE
              write(10,*)L1,L2,Lgrid(3)
#else
              write(10)L1,L2,Lgrid(3)
#endif
           endif
! write data plane by plane
           do local_z=0,Lgrid_local(3)-1
              global_z = local_z + Lgrid_z_start

              plane=0.0

              do iy=0,Lgrid(2)-1
                 do ix=0,Lgrid(1)-1

!                    print '(5(1x,i5))',proc_id,ix,iy,ix/ResamplingRate,iy/ResamplingRate
                    plane(ix/ResamplingRate,iy/ResamplingRate)= &
                         plane(ix/ResamplingRate,iy/ResamplingRate)+density(ix,iy,local_z)
                 enddo
              enddo
              plane=plane/dble(ResamplingRate**2)


#ifdef ASCII_WRITE
              write(10,*) global_z
              write(10,'(8(1x,e13.6))') plane
#else
              write(10) global_z
              write(10) plane
#endif
           enddo

           close(10)

        endif

! MPI barrier
        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
     enddo

     deallocate(plane)

! exit
     info = ' written '//trim(filename)
     call message(trim(info))
 
!  enddo


  endif


  return
end subroutine write_density


character(len=4) function write_string(i)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::i

  write(write_string,'(i4)')i
  if (i<=9) then 
     write_string(1:3)='000'
  else if (i<=99) then
     write_string(1:2)='00'
  else if (i<=999) then
     write_string(1:1)='0'
  endif
  return
end function write_string
  
! ****************
! MPI I/O routines
! ****************

subroutine message(messg)
  use pinocchio_module
  implicit none

  character(*) :: messg
  
  if(proc_id .eq. 0) then 
     write (*,'(a)') trim(messg)
#ifndef NO_FLUSH
     call flush(6)
#endif
  endif

  return
end subroutine message
      
subroutine abort_code(messg)
  use pinocchio_module
  implicit none
  !    
  character(*) :: messg
  
  write (*,'(a)') trim(messg)
  write (*,*) ' MPI_ABORT called by processor: ',proc_id
#ifndef NO_FLUSH
  call flush(6)
#endif
  call MPI_ABORT(MPI_COMM_WORLD,ierr)
  call MPI_FINALIZE(' .... stop ')
  stop
end subroutine abort_code

subroutine finish(messg)
  use pinocchio_module
  implicit none
  !    
  character(*) :: messg
  
  call message(messg)

  call MPI_FINALIZE(ierr)

 stop
end subroutine finish


subroutine write_snapshot(z)

! writes displacements of all particles in a GADGET format
! each processor writes one snapshot file???

  use pinocchio_module
  use fmax_grids
  use fmax_main_memory
  implicit none

  real(kind=doubleR),intent(in)   :: z
  real(kind=doubleR),external     :: fomega,Hubble,pbc,grow
  integer(kind=singleI),parameter :: unit=10

  real(kind=doubleR)    :: vf, Time
  integer(kind=singleI) :: i,j,k,kk, NPart, NPartTot1, NPartTot2
  character(len=300)    :: filename
  character(len=10)     :: label
  integer(kind=doubleI),allocatable :: IDs(:,:,:)
  real(kind=singleR),allocatable    :: posvel(:,:,:,:)

  vf = InterPartDist*fomega(z)*Hubble(z)/(1.+z)**0.5       ! GADGET format requres velocities to be divided by sqrt(a)
  Time=1./(1+z)

  write(info,'("velocity factor=",g15.8)')vf
  call message(info)

!  NPartTot1 = iand(dble(Lgrid(1))*dble(Lgrid(2))*dble(Lgrid(3)),dble(2**8)-1)
!  NPartTot2 = int(dble(Lgrid(1))*dble(Lgrid(2))*dble(Lgrid(3))/2**8)
  NPartTot1 = Lgrid(1)*Lgrid(2)*Lgrid(3)
  NPartTot2 = 0
  NPart  = Lgrid_local(1)*Lgrid_local(2)*Lgrid_local(3)

  write(label,'(f10.4)')z
  filename='Data/displacements.'//trim(adjustl(label))//'.'//trim(run)//'.'
  write(label,'(i10)')proc_id
  filename=trim(filename)//trim(adjustl(label))

  print *,proc_id,label,trim(filename)

  write(*,'("Processor ",i4," writing file ",a)')proc_id,trim(filename)
  open(unit,file=filename,status='unknown',form='unformatted')

  write(unit)'HEAD',256
  write(unit)0,NPart,0,0,0,0, &                       ! NPart[6]
       0.0d0,Mpart*Hubble100*1e-10,0.0d0,0.0d0,0.0d0,0.0d0, &      ! Mass[6]
       Time,z, &                                      ! Time, Redshift
       0,0,    &                                      ! flag_sfr, flag_feedback
       0,NPartTot1,0,0,0,0, &                         ! NPartTotal[6]
       0,nproc, &                                     ! flag_cooling, num_files
       BoxSize(1)*Hubble100,Omega0,OmegaLambda,Hubble100, &  ! BoxSize, Omega0, OmegaLambda, HubbleParam
       0,0, &                                         ! flag_stellarage, flag_metals
       0,NPartTot2,0,0,0,0, &                         ! npartTotalHighWord[6]
       0,0,0, &                                       ! flag_entropy_instead_u, flag_metalcooling, flag_stellarevolution
       '                                                    '

  total_allocated = total_allocated + 4*3*npart_local
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating posvel...')  
  allocate(posvel(3,0:Lgrid_local(1)-1,0:Lgrid_local(2)-1,0:Lgrid_local(3)-1))
  if (alloc_verbose) then
     call message ('allocated posvel')  
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  do kk=0,Lgrid_local(3)-1
     k=kk+Lgrid_z_start
     do j=0,Lgrid_local(2)-1
        do i=0,Lgrid_local(1)-1

           posvel(1,i,j,kk)=pbc((i+VEL_for_displacements(1,i,j,kk))*InterPartDist,BoxSize(1))*Hubble100
           posvel(2,i,j,kk)=pbc((j+VEL_for_displacements(2,i,j,kk))*InterPartDist,BoxSize(2))*Hubble100
           posvel(3,i,j,kk)=pbc((k+VEL_for_displacements(3,i,j,kk))*InterPartDist,BoxSize(3))*Hubble100

        enddo
     enddo
  enddo

  write(unit)'POS ',int(NPart*3*4)
  write(unit)posvel

  posvel=VEL_for_displacements*vf

  write(unit)'VEL ',int(NPart*3*4)
  write(unit)posvel

  deallocate(posvel)
  total_allocated = total_allocated - 4*3*npart_local
  if (alloc_verbose) call message ('de-allocated posvel')  

  total_allocated = total_allocated + 8*npart_local
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating IDs...')
  allocate(IDs(0:Lgrid_local(1)-1,0:Lgrid_local(2)-1,0:Lgrid_local(3)-1))
  if (alloc_verbose) then
     call message ('allocated IDs')  
     write (info,'('' total allocated memory: '',f8.3,'' Gbytes'')') total_allocated/Gbyte
     call message(trim(info))
  endif

  do kk=0,Lgrid_local(3)-1
     k=kk+Lgrid_z_start
     do j=0,Lgrid_local(2)-1
        do i=0,Lgrid_local(1)-1
           IDs(i,j,kk)=dble(i)+(dble(j)+dble(k)*dble(Lgrid(2)))*dble(Lgrid(1))
        enddo
     enddo
  enddo

  write(unit)'ID  ',int(NPart*8)
  write(unit)IDs

  total_allocated = total_allocated - 8*npart_local
  deallocate(IDs)
  if (alloc_verbose) call message ('de-allocated IDs')  

  close(unit)

  return
end subroutine write_snapshot


!!!subroutine read_fmax()
!!!
!!!  use pinocchio_module
!!!  use fmax_grids
!!!  use fmax_main_memory
!!!  use final_products
!!!  implicit none
!!!
!!!! local variables
!!!  integer(kind=singleI)    :: iproc, local_z, global_z, Li, Lj, Lk
!!!  character(len=80)        :: filename
!!!  logical                  :: flag
!!!
!!!  write(filename,"('Data/Fmax.',a,'.d')")trim(StartFlag)
!!!
!!!  do iproc=0,nproc-1
!!!     if(iproc .eq. proc_id) then
!!!
!!!        if (proc_id==0) then
!!!           inquire(file=filename,exist=flag)
!!!           if (.not.flag) then
!!!              write(info,"('ERROR IN READING FMAX: file ',a,' not found')")trim(filename)
!!!              call message(info)
!!!              call abort_code('')
!!!           endif
!!!        endif
!!!
!!!#ifdef ASCII_WRITE
!!!        open(10,file=filename,status="old")
!!!#else
!!!        open(10,file=filename,status="old",form='unformatted')
!!!#endif
!!!
!!!#ifdef ASCII_WRITE
!!!        read(10,*)Li,Lj,Lk
!!!#else
!!!        read(10)Li,Lj,Lk
!!!#endif
!!!        if (proc_id==0) then
!!!           if (Li /= Lgrid(1) .or. Lj /= Lgrid(2) .or. Lk /= Lgrid(3)) then
!!!              write(info,"('ERROR IN READING FMAX: mismatched grid sizes',3(1x,i9))")Li,Lj,Lk
!!!              call message(info)
!!!              call abort_code('')
!!!           endif
!!!        endif
!!!
!!!! skip other planes
!!!        do global_z=0,Lgrid_z_start-1
!!!#ifdef ASCII_WRITE
!!!           read(10,*) Li
!!!           read(10,*) Fmax(:,:,0)
!!!#else
!!!           read(10) Li
!!!           read(10) 
!!!#endif
!!!
!!!           if (Li /= global_z) then
!!!              write(info,"('ERROR IN READING FMAX: mismatched plane',3(1x,i9))")Li,global_z
!!!              call message(info)
!!!              call abort_code('')
!!!           endif
!!!
!!!        enddo
!!!
!!!! reading planes
!!!        do local_z=0,Lgrid_local(3)-1
!!!           global_z = local_z + Lgrid_z_start
!!!
!!!#ifdef ASCII_WRITE
!!!           read(10,*) Li
!!!           read(10,*) Fmax(:,:,local_z)
!!!#else
!!!           read(10) Li
!!!           read(10) Fmax(:,:,local_z)
!!!#endif
!!!
!!!           if (Li /= global_z) then
!!!              write(info,"('ERROR IN READING FMAX: mismatched plane',3(1x,i9))")Li,global_z
!!!              call message(info)
!!!              call abort_code('')
!!!           endif
!!!
!!!        enddo
!!!
!!!! close file
!!!        close(10)
!!!
!!!     endif
!!!
!!!! MPI barrier
!!!     call MPI_BARRIER(MPI_COMM_WORLD,ierr)
!!!  enddo
!!!
!!!! exit
!!!  info = ' read file '//trim(filename)
!!!  call message(trim(info))
!!!  return
!!!end subroutine read_fmax
!!!
!!!
!!!subroutine read_vmax()
!!!
!!!  use pinocchio_module
!!!  use fmax_grids
!!!  use fmax_main_memory
!!!  use final_products
!!!  implicit none
!!!
!!!! local variables
!!!  integer(kind=singleI)    :: iproc, local_z, global_z,ia, Li, Lj, Lk
!!!  character(len=80)        :: filename
!!!  logical                  :: flag
!!!
!!!  do ia=1,3
!!!
!!!     if (ia==1) then
!!!        write(filename,"('Data/Velx.',a,'.d')")trim(StartFlag)
!!!     else if (ia==2) then
!!!        write(filename,"('Data/Vely.',a,'.d')")trim(StartFlag)
!!!     else if (ia==3) then
!!!        write(filename,"('Data/Velz.',a,'.d')")trim(StartFlag)
!!!     endif
!!!
!!!     do iproc=0,nproc-1
!!!        if(iproc .eq. proc_id) then
!!!
!!!           if (proc_id==0) then
!!!              inquire(file=filename,exist=flag)
!!!              if (.not.flag) then
!!!                 write(info,"('ERROR IN READING VMAX: file ',a,' not found')")trim(filename)
!!!                 call message(info)
!!!                 call abort_code('')
!!!              endif
!!!           endif
!!!
!!!!           print *,'read_vmax, Task',proc_id,'opening file',trim(filename)
!!!
!!!#ifdef ASCII_WRITE
!!!           open(10,file=filename,status="old")
!!!#else
!!!           open(10,file=filename,status="old",form='unformatted')
!!!#endif
!!!
!!!#ifdef ASCII_WRITE
!!!           read(10,*)Li,Lj,Lk
!!!#else
!!!           read(10)Li,Lj,Lk
!!!#endif
!!!           if (proc_id==0) then
!!!              if (Li /= Lgrid(1) .or. Lj /= Lgrid(2) .or. Lk /= Lgrid(3)) then
!!!                 write(info,"('ERROR IN READING VMAX: mismatched grid sizes',3(1x,i9))")Li,Lj,Lk
!!!                 call message(info)
!!!                 call abort_code('')
!!!              endif
!!!           endif
!!!
!!!! skip other planes
!!!           do global_z=0,Lgrid_z_start-1
!!!#ifdef ASCII_WRITE
!!!              read(10,*) Li
!!!              read(10,*) vmax(ia,:,:,0)
!!!#else
!!!              read(10) Li
!!!              read(10)
!!!#endif
!!!
!!!!              print *,'read_vmax, Task',proc_id,'skipping plane',Li
!!!
!!!              if (Li /= global_z) then
!!!                 write(info,"('ERROR IN READING VMAX: mismatched plane',3(1x,i9))")Li,global_z
!!!                 call message(info)
!!!                 call abort_code('')
!!!              endif
!!!
!!!           enddo
!!!
!!!! reading planes
!!!           do local_z=0,Lgrid_local(3)-1
!!!              global_z = local_z + Lgrid_z_start
!!!
!!!#ifdef ASCII_WRITE
!!!              read(10,*) Li
!!!              read(10,*) vmax(ia,:,:,local_z)
!!!#else
!!!              read(10) Li
!!!              read(10) vmax(ia,:,:,local_z)
!!!#endif
!!!
!!!!              print *,'read_vmax, Task',proc_id,'reading plane',Li
!!!
!!!              if (Li /= global_z) then
!!!                 write(info,"('ERROR IN READING VMAX: mismatched plane',3(1x,i9))")Li,global_z
!!!                 call message(info)
!!!                 call abort_code('')
!!!              endif
!!!
!!!           enddo
!!!
!!!! close file
!!!           close(10)
!!!
!!!        endif
!!!
!!!! MPI barrier
!!!        call MPI_BARRIER(MPI_COMM_WORLD,ierr)
!!!     enddo
!!!
!!!     info = ' read file '//trim(filename)
!!!     call message(trim(info))
!!!
!!!  enddo
!!!
!!!! exit
!!!  return
!!!end subroutine read_vmax
!!!
!!!
!!!subroutine read_variances()
!!!
!!!  use pinocchio_module
!!!  use fmax_smoothing_radii
!!!  implicit none
!!!
!!!  integer(kind=singleI)::ismooth
!!!
!!!  if (proc_id==0) then
!!!
!!!     open(10,file=trim(StartFlag)//'.var',status='old')
!!!     do ismooth=1,nsmooth
!!!        read(10,*) Smoothing_Radii(ismooth) &
!!!          , Smoothing_Variances(ismooth) &
!!!          , True_Variances(ismooth)
!!!     enddo
!!!
!!!  endif
!!!
!!!  info='read file '//trim(StartFlag)//'.var'
!!!  call message(info)
!!!
!!!  call MPI_BCAST(Smoothing_Radii,     nsmooth, MPI_DOUBLE_PRECISION, 0, MPI_COMM_WORLD, ierr)
!!!  call MPI_BCAST(Smoothing_Variances, nsmooth, MPI_DOUBLE_PRECISION, 0, MPI_COMM_WORLD, ierr)
!!!  call MPI_BCAST(True_Variances,      nsmooth, MPI_DOUBLE_PRECISION, 0, MPI_COMM_WORLD, ierr)
!!!
!!!  return
!!!end subroutine read_variances
