!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#define OUTPUT_PARAMETERS
!#define READ_PARAMS_FROM_FILE
!#define WRITE_GROUPIDS
#define ROTATE_BOX

subroutine fragment()

  use fragment_run_flags
  use pinocchio_module
  implicit none
  character(len=24), external :: fdate

! timing
  cputime(10)=MPI_WTIME()

  write(info,'("[",a24,"] Second part: fragmentation")') fdate()
  call message(info)

! Initializes everything
  call init_fragment()

  cputime(8)=MPI_WTIME()
! Generates the group catalogue
  call groups()
  cputime(8)=MPI_WTIME()-cputime(8)

! final cpu time
  cputime(10) = MPI_WTIME() - cputime(10)
  write(info,'(''['',a24,''] Finishing fragment, total cputime = '',g14.6)') fdate(),cputime(10)
  call message (trim(info))

  return
end subroutine fragment


subroutine init_fragment()

! Initializes all quantities for the run

  use pinocchio_module
  use fmax_smoothing_radii
  use fragment_grids
  use fragment_run_flags
  use fragment_virial_params
  use fragment_matrices
  use fragment_group_vectors

  implicit none

  integer(kind=singleI)::i,ngood,Lgrid12,iz,j,k,kk,tot_good
  real(kind=doubleR)::var_grid
  real(kind=doubleR),external::grow  !,NumberDensity
  character(len=24),external :: fdate

#ifdef READ_PARAMS_FROM_FILE
  real(kind=doubleR)::f_amin,f_amax
#endif

! re-computes mass variances with top-hat smoothing
  filter=2
  call Lambda_start()

! inverse collapse times for the requested outputs
  do i=1,nout
     fout(i)=1./grow(zout(i))
  enddo

! ***************************************


  var_grid=True_variances(nsmooth)

#ifndef READ_PARAMS_FROM_FILE
  f_m  = 0.35
  f_rm = 0.70
  f_s  = 0.06
  f_a  = 0.285 !+ (log10(var_grid/InterPartDist) - 0.36) * 0.11
!  f_a  = min (f_a,0.26d0)
!  f_a  = max (f_a,0.22d0)
  f_ra = 0.18 ! - (f_a - 0.22) * 3.5
  espo=1.7
#else
  open(10,file='params.'//trim(run),status='old')
  read(10,*)f_m
  read(10,*)f_rm
  read(10,*)f_s
  read(10,*)f_a
!  read(10,*)f_amax
  read(10,*)f_ra
  read(10,*)espo

!  f_a  = f_amin + (log10(var_grid/InterPartDist) - 0.36) * 0.11
!  f_a  = min (f_a,f_amax)
!  f_a  = max (f_a,f_amin)
!  f_ra = f_ra - (f_a - f_amin) * 3.5
#endif

! writes information on the run
     
#ifdef OUTPUT_PARAMETERS
  if (proc_id==0) then
     write(*,'("Fragmentation parameters:")')
     write(*,'(a,10(3x,f6.3))')'f_m   = ',f_m
     write(*,'(a,10(3x,f6.3))')'f_a   = ',f_a
     write(*,'(a,10(3x,f6.3))')'f_rm  = ',f_rm
     write(*,'(a,10(3x,f6.3))')'f_ra  = ',f_ra
     write(*,'(a,10(3x,f6.3))')'f_s   = ',f_s
     write(*,'(a,10(3x,f6.3))')'espo  = ',espo
     write(*,'(a,10(3x,f6.3))')'sig2  = ',var_grid  !/InterPartDist
     write(*,*)
  endif
#endif


! this routine redistributes Fmax and vmax from planes to sub-boxes
  cputime(6)=MPI_WTIME()
  write(info,'(''['',a24,''] Starting re-distribution of Fmax'')') fdate()
  call message (trim(info))

  call fragment_redistribute_big_matrices()

  cputime(6)=MPI_WTIME()-cputime(6)
  write(info,'(''['',a24,''] Re-distribution of Fmax done, total cputime = '',g14.6)') fdate(),cputime(6)
  call message (trim(info))

! Sorts the Fmax_frag matrix
  cputime(7)=MPI_WTIME()
  write(info,'("[",a24,"] sorting...")')fdate()
  call message(trim(info))

  call index_sort(Npart_frag,Fmax_frag,indices)

  cputime(7)=MPI_WTIME()-cputime(7)
  write(info,'("[",a24,"] sorting done, total cpu time = ",g14.6)')fdate(),cputime(7)
  call message(trim(info))

! counts the number of Fmax peaks down to the final redshift
! to know the total number of groups
  ngroups_tot=0    ! number of groups, group 1 is the filament group
  ngood=0          ! number of groups out of the safety boundary
  Lgrid12=Lgrid_frag(1)*Lgrid_frag(2)
  do iz=Npart_frag,1,-1

! position on the local box
!     k=1+(indices(iz)-1)/Lgrid12
!     kk=indices(iz)-1-(k-1)*Lgrid12
!     j=1+kk/Lgrid_frag(1)
!     i=1+(kk-(j-1)*Lgrid_frag(1))

! position on the local box
     k=1+(iz-1)/Lgrid12
     kk=iz-1-(k-1)*Lgrid12
     j=1+kk/Lgrid_frag(1)
     i=1+(kk-(j-1)*Lgrid_frag(1))

! avoid borders
     if (i==1 .or. i==Lgrid_frag(1) .or. &
         j==1 .or. j==Lgrid_frag(2) .or. &
         k==1 .or. k==Lgrid_frag(3)) cycle

! peak condition
!     if ( Fmax_frag(indices(iz)) >= fout(nout) .and. &
!          (i==Lgrid_frag(1)-1 .or. Fmax_frag(indices(iz)) > Fmax_frag(i+1+(j-1)*Lgrid_frag(1)+(k-1)*Lgrid12)) .and. &
!          (i==2               .or. Fmax_frag(indices(iz)) > Fmax_frag(i-1+(j-1)*Lgrid_frag(1)+(k-1)*Lgrid12)) .and. &
!          (j==Lgrid_frag(2)-1 .or. Fmax_frag(indices(iz)) > Fmax_frag(i  +(j  )*Lgrid_frag(1)+(k-1)*Lgrid12)) .and. &
!          (j==2               .or. Fmax_frag(indices(iz)) > Fmax_frag(i  +(j-2)*Lgrid_frag(1)+(k-1)*Lgrid12)) .and. &
!          (k==Lgrid_frag(3)-1 .or. Fmax_frag(indices(iz)) > Fmax_frag(i  +(j-1)*Lgrid_frag(1)+(k  )*Lgrid12)) .and. &
!          (k==2               .or. Fmax_frag(indices(iz)) > Fmax_frag(i  +(j-1)*Lgrid_frag(1)+(k-2)*Lgrid12))) then

! peak condition
     if ( Fmax_frag(iz) >= fout(nout) .and. &
          (i==Lgrid_frag(1)-1 .or. Fmax_frag(iz) > Fmax_frag(i+1+(j-1)*Lgrid_frag(1)+(k-1)*Lgrid12)) .and. &
          (i==2               .or. Fmax_frag(iz) > Fmax_frag(i-1+(j-1)*Lgrid_frag(1)+(k-1)*Lgrid12)) .and. &
          (j==Lgrid_frag(2)-1 .or. Fmax_frag(iz) > Fmax_frag(i  +(j  )*Lgrid_frag(1)+(k-1)*Lgrid12)) .and. &
          (j==2               .or. Fmax_frag(iz) > Fmax_frag(i  +(j-2)*Lgrid_frag(1)+(k-1)*Lgrid12)) .and. &
          (k==Lgrid_frag(3)-1 .or. Fmax_frag(iz) > Fmax_frag(i  +(j-1)*Lgrid_frag(1)+(k  )*Lgrid12)) .and. &
          (k==2               .or. Fmax_frag(iz) > Fmax_frag(i  +(j-1)*Lgrid_frag(1)+(k-2)*Lgrid12))) then

        ngroups_tot=ngroups_tot+1

        if ( i>safe .and. i<=Lgrid_frag(1)-safe .and. &
             j>safe .and. j<=Lgrid_frag(2)-safe .and. &
             k>safe .and. k<=Lgrid_frag(3)-safe) then
           ngood=ngood+1
        endif
     endif
  enddo

  call MPI_REDUCE(ngood, tot_good, 1, MPI_INTEGER, MPI_SUM, 0, MPI_COMM_WORLD, ierr)

  write(info,'("[",a24,"] Processor ",i4," found ",i9," peaks, excluding safety belt: ",i9)') fdate(),proc_id,ngroups_tot,ngood
  call message (trim(info))

  write(info,'("Total number of peaks: ",i9)')tot_good
  call message(trim(info))

  ngroups_tot=ngroups_tot+2

  call fragment_allocate_group_vectors()

! bye!

  return
end subroutine init_fragment


subroutine groups()

! Fragment the collapsed medium into groups

  use pinocchio_module
  use fragment_grids
  use fragment_run_flags
  use fragment_merger
  use fragment_link
  use fragment_group_vectors
  use fragment_matrices 
#ifdef PLC
  use tocompute_PLC_condition
#endif
  implicit none

  integer(kind=singleI),parameter::nv=6
  integer(kind=singleI),dimension(nv,nv)::merge
  integer(kind=singleI),dimension(nv)::neigh
  integer(kind=singleI),dimension(3,nv)::fil_list
  integer(kind=singleI),dimension(6)::nm
  integer(kind=singleI)::NMERG,iout,Lgrid12,nn,i,j,k,ifil,indx,neigrp
  integer(kind=singleI)::iz,kk,i1,i2,j1,j2,k1,k2
  integer(kind=singleI)::i_global,j_global,k_global
  integer(kind=singleI)::ig3,small,large
  integer(kind=singleI)::nf,to_group,accgrp,ig1,ig2
  integer(kind=singleI)::accrflag,nstep,nmerge,nstep_p,perc
  real(kind=doubleR)::ratio,best_ratio,d2,r2
  logical::mask(-1:1,-1:1,-1:1)
  logical::merge_flag,good_particle
  real(kind=doubleR),external::inverse_grow
  integer(kind=singleI),dimension(0:13)::counter,all_counters
! counters:
!  0 number of peaks
!  i number of particles with i neighbours (i<=6)
!  7 number of accretion events
!  8 number of accretion events before checking a merger
!  9 number of accretion events after checking a merger
! 10 number of merging events
! 11 number of major mergers
! 12 number of filament particles
! 13 number of accreted filament particles

  character(len=24), external :: fdate

#ifdef PLC
  real(kind=doubleR),external::condition_PLC,zbrent,proper_dist
  real(kind=doubleR)::aa,bb,Fplc
  integer(kind=singleI),dimension(3,nv)::check_again
  integer(kind=singleI)::ir,jr,kr
#endif 

! Initializations

  NMERG=0                   ! NUMBER OF MAJOR MERGER EVENTS
  ngroups=1                 ! number of groups + filaments
  Lgrid12=Lgrid_frag(1)*Lgrid_frag(2)
  iout=1

  mass=0     ; point=0     ; bottom=0
  GR_LL=-1   ; HALO_APP=-1  !; MCENTR=-1
  MASS_AT_MERGER=0  ;  MERGED_WITH=-1
  MAJORZ=-1. ; T_ACC=-1.   ; T_APPEAR=-1.
  centres=0. ; velgrp=0.   
  group_ID=0 ; link_list=0
  nm=0

  counter=0 ; all_counters=0

  point(1)=Npart_frag+1        ! filaments are not grouped!
  bottom(1)=Npart_frag+1

! This defines the number of neighbours considered

  if (nv.eq.26) then
! In this case all the 26 neighbouring points are considered
! (the central one is obviously excluded)
     mask=.true.
     mask(0,0,0)=.false.

  else if (nv.eq.6) then
! In this case considers only 6 neighbouring points
     mask=.false.
     mask(-1, 0, 0) =.true.
     mask( 1, 0, 0) =.true.
     mask( 0,-1, 0) =.true.
     mask( 0, 1, 0) =.true.
     mask( 0, 0,-1) =.true.
     mask( 0, 0, 1) =.true.
  else
     stop 'wrong nv'
  endif


! ********************************************
  write(info,'("[",a24,"] Starting the fragmentation process")')fdate()
  call message(trim(info))
! ********************************************


! Calculates the number of steps required

  iz=Npart_frag
  
  do while (Fmax_frag(indices(iz)) >= fout(nout))
     iz=iz-1
  enddo
  nstep=Npart_frag-iz
  nstep_p=nstep/20

! ************************************************************************
! START OF THE CYCLE ON POINTS
! ************************************************************************
  do iz=Npart_frag,1,-1

! Outputs if relevant
     do while (Fmax_frag(indices(iz)) < fout(iout))

        write(info,'("[",a24,"] Writing output at z=",f8.4)')fdate(), zout(iout)
        call message(trim(info))

        call write_cat(iout)

        call compute_mf(iout)

#ifdef PLC
        if (iout==nout .and. Flimit_plc>0.) call write_plc()
#endif

        if (iout==nout) goto 99

        iout=iout+1

     end do

! Fraction of steps done
     k=Npart_frag-iz+1
     perc=k/(nstep_p)*5

     if (mod(k,nstep_p).eq.0) then
        write(info,'("[",a24,"] *** ",i3,"% Done, F =",f6.2,",  z = ",f6.2)') fdate(),&
             perc,Fmax_frag(indices(iz)),inverse_grow(1./Fmax_frag(indices(iz)))

        call message(trim(info))

     endif

! More initializations

     neigrp=0               ! number of neighbouring groups
     nf=0                   ! number of neighbouring filament points
     accrflag=0             ! if =1 all the neighbouring filaments
                            !   are accreted 
     neigh=0                ! neighbours

! grid coordinates from the indices (frag-box coordinates)

     k=1+(indices(iz)-1)/Lgrid12
     kk=indices(iz)-1-(k-1)*Lgrid12
     j=1+kk/Lgrid_frag(1)
     i=1+(kk-(j-1)*Lgrid_frag(1))

! skips if the point is at the border

     if (i==1 .or. i==Lgrid_frag(1)) cycle
     if (j==1 .or. j==Lgrid_frag(2)) cycle
     if (k==1 .or. k==Lgrid_frag(3)) cycle

     i_global=i+nstart(1)
     j_global=j+nstart(2)
     k_global=k+nstart(3)

     good_particle=(i>safe .and. i<=Lgrid_frag(1)-safe .and. j>safe .and. j<=Lgrid_frag(2)-safe .and. k>safe .and. k<=Lgrid_frag(3)-safe)

! List of neighbouring groups
     nn=0
     do k1=-1,+1
        do j1=-1,+1
           do i1=-1,+1
              if (mask(i1,j1,k1)) then
                 i2=i+i1
                 j2=j+j1
                 k2=k+k1
                 nn=nn+1
                 neigh(nn)=group_ID(i2+(j2-1)*Lgrid_frag(1)+(k2-1)*Lgrid12)
! list of touching points that are in filaments (not counted as groups)
! NB: IT ASSUMES THAT GROUP 1 CONTAINS FILAMENTS
                 if (neigh(nn)<=1 .and. neigh(nn)/=0) then
                    nf=nf+1
                    neigh(nn)=0
                    fil_list(1,nf)=i2
                    fil_list(2,nf)=j2
                    fil_list(3,nf)=k2
                 endif
              endif
           enddo
        enddo
     enddo

! Cleans the list of neighbouring groups
     call clean_list(nv,neigh)

! Number of neighbouring groups
     neigrp=0
     do while (neigh(neigrp+1)>1) 
        neigrp=neigrp+1
        if (neigrp==nv) exit
     enddo

     if (neigrp>0 .and. good_particle) counter(neigrp)=counter(neigrp)+1


!     write(190+proc_id,'(f12.5,99(1x,i6))') &
!          Fmax_frag(indices(iz)),i_global,j_global,k_global,neigrp,neigh(1:neigrp)


#ifdef PLC
! Past light cone construction:
! checks whether one of the neighbouring groups has just crossed the PLC

     ! QUESTO BISOGNA CAPIRE SE FUNZIONA (E SE NON FUNZIONA PERCHE' NO)
     !dist_at_F=proper_dist(inverse_grow(1./Fmax_frag(indices(iz))))/InterPartDist

     ! is it time to reconstruct the PLC?
     if (Fmax_frag(indices(iz))<Flimit_plc) then

        check_again=999
        do ig1=1,neigrp

           ! is the group good and massive enough?
           if (good(neigh(ig1)) .and. mass(neigh(ig1))>=MIN_MASS_FOR_CAT) then

              thisgroup=neigh(ig1)

              ! loop on replications
              do ir=-number_of_rep,number_of_rep
                 replicate(1)=ir
                 do jr=-number_of_rep,number_of_rep
                    replicate(2)=jr
                    do kr=-number_of_rep,number_of_rep
                       replicate(3)=kr

                       bb=condition_PLC(Fmax_frag(indices(iz)))

                       ! in this unlikely case we catch the group just on the PLC
                       if (bb==0.) then
                          call store_plc(Fmax_frag(indices(iz)))

                       else if (bb>0.) then

                          ! if it is outside the PLC then check whether it has just passed
                          aa=condition_PLC(Flast(neigh(ig1)))

                          if (aa<0.0) then

                             ! in this case the group has passed through the PLC since last time
                             Fplc=zbrent(condition_PLC,Flast(neigh(ig1)),Fmax_frag(indices(iz)),1.d-6)

                             call store_plc(Fplc)
                          endif

                       else if (bb>-3.*InterPartDist) then  ! TEST

                          ! in this case the group is near the PLC and it will be checked again
                          check_again(:,ig1)=replicate

                       endif

                    end do
                 end do
              end do

           end if

           ! updates the Flast vector
           Flast(neigh(ig1))=Fmax_frag(indices(iz))
        end do

     else if (Flimit_plc>0.) then

        do ig1=1,neigrp
           Flast(neigh(ig1))=Fmax_frag(indices(iz))
        enddo

     endif
#endif

! Is the point isolated?

     if (neigrp==0 .and. nf==0) then
! **********************************************************************
!                       FIRST CASE: PEAK
! **********************************************************************

! New group
        if (good_particle) counter(0)=counter(0)+1
        ngroups=ngroups+1
        if (ngroups > ngroups_tot) then

           write(info,'("TOO MANY GROUPS???",i10,1x,i10,1x,i10)')proc_id,ngroups,ngroups_tot
           call abort_code(info)

        endif
        centres(1,ngroups)=i_global-.5
        centres(2,ngroups)=j_global-.5
        centres(3,ngroups)=k_global-.5
        velgrp(1,ngroups)=vx(indices(iz))
        velgrp(2,ngroups)=vy(indices(iz))
        velgrp(3,ngroups)=vz(indices(iz))
#ifdef PLC
        if (Flimit_plc>0.) Flast(ngroups)=Fmax_frag(indices(iz))
#endif
        mass(ngroups)=1
        name(ngroups)=i_global+(j_global-1)*Lgrid(1)+(k_global-1)*Lgrid(1)*Lgrid(2)
        good(ngroups)=(i>safe .and. i<=Lgrid_frag(1)-safe .and. j>safe .and. j<=Lgrid_frag(2)-safe .and. k>safe .and. k<=Lgrid_frag(3)-safe)
        group_ID(indices(iz))=ngroups
        point(ngroups)=indices(iz)
        bottom(ngroups)=indices(iz)
        link_list(indices(iz))=indices(iz)
        GR_LL(NGROUPS)=group_ID(indices(iz))
        HALO_APP(NGROUPS)=NGROUPS


     else if (neigrp.eq.1) then
! **********************************************************************
!                       SECOND CASE: 1 GROUP
! **********************************************************************

! if the points touches only one group, check whether to accrete the 
! point on it

        call accr_cond(i_global,j_global,k_global,indices(iz) &
             ,Fmax_frag(indices(iz)),neigh(1),d2,r2)

!        write(120+proc_id,*)Fmax_frag(indices(iz)),d2,r2

        if (d2.lt.r2) then

! *********************
! accretion on a group!
! *********************

           if (good_particle) counter(7)=counter(7)+1
           accrflag=1
           to_group=neigh(1)
           call accretion(to_group,i_global,j_global,k_global,indices(iz),Fmax_frag(indices(iz)))

        else

! *********
! filament!
! *********

           mass(1)=mass(1)+1
           group_ID(indices(iz))=1
           link_list(indices(iz))=indices(iz)

        endif

     else if (neigrp.gt.1) then
! **********************************************************************
!                       THIRD CASE: >1 GROUP
! **********************************************************************

! In this case the point touches more than one group

! *********************
! accretion on a group?
! *********************

        best_ratio=(10.*Lgrid_frag(1))**2
        accgrp=0
        do ig1=1,neigrp
           call accr_cond(i_global,j_global,k_global,indices(iz) &
                ,Fmax_frag(indices(iz)),neigh(ig1),d2,r2)

!           write(120+proc_id,*)Fmax_frag(indices(iz)),d2,r2

           ratio=d2/r2
           if (ratio.lt.1..and.ratio.lt.best_ratio) then
              best_ratio=ratio
              accgrp=ig1
           endif
        enddo

        if (accgrp.gt.0) then
           if (good_particle) counter(7)=counter(7)+1
           if (good_particle) counter(8)=counter(8)+1
           accrflag=1
           to_group=neigh(accgrp)
           call accretion(neigh(accgrp),i_global,j_global,k_global,indices(iz),Fmax_frag(indices(iz)))
        endif

! Then checks whether the groups must be merged together

        nmerge=0
        do ig1=1,neigrp
           do ig2=1,ig1-1
              merge(ig1,ig2)=0
              call merge_cond(Fmax_frag(indices(iz)),neigh(ig1),neigh(ig2), &
                   merge_flag)
              if (merge_flag) then
                 merge(ig1,ig2)=1
                 nmerge=nmerge+1
              endif
           enddo
        enddo

! ******************
! merging of groups!
! ******************

! The group number of the largest group is preserved

        if (nmerge.gt.0) then
           do ig1=1,neigrp
              do ig2=1,ig1-1
                 if (merge(ig1,ig2).eq.1.and.neigh(ig1).ne.neigh(ig2)) then
                    if (good_particle) counter(10)=counter(10)+1
                    if (mass(neigh(ig1)).gt.mass(neigh(ig2))) then
                       call merge_groups(neigh(ig1),neigh(ig2),Fmax_frag(indices(iz)))
                       large=neigh(ig1)
                       small=neigh(ig2)
                       if (to_group.eq.small) to_group=large
                       do ig3=1,neigrp
                          if (neigh(ig3).eq.small) neigh(ig3)=large
                       enddo
                    else
                       call merge_groups(neigh(ig2),neigh(ig1),Fmax_frag(indices(iz)))
                       small=neigh(ig1)
                       large=neigh(ig2)
                       if (to_group.eq.small) to_group=large
                       do ig3=1,neigrp
                          if (neigh(ig3).eq.small) neigh(ig3)=large
                       enddo
                    endif
                    if (mass(large)<5*mass(small) .and. good_particle) counter(11)=counter(11)+1
                 endif
              enddo
           enddo
        endif

! If relevant, it tries again to accrete the point

        if (accgrp.eq.0) then
#ifndef PLC
           call clean_list(nv,neigh)
#else
           if (Flimit_plc>0.) then
              call clean_list_with_PLC(nv,neigh,check_again)
           else
              call clean_list(nv,neigh)
           endif
#endif

           neigrp =0
           do while(neigh(neigrp+1).gt.1)
              neigrp = neigrp + 1
              if (neigrp==nv) exit
           enddo

           accgrp=0
           best_ratio=(10.*Lgrid_frag(1))**2
           do ig1=1,neigrp
              call accr_cond(i_global,j_global,k_global,indices(iz) &
                   ,Fmax_frag(indices(iz)),neigh(ig1),d2,r2)

!              write(120+proc_id,*)Fmax_frag(indices(iz)),d2,r2

              ratio=d2/r2
              if (ratio.lt.best_ratio) then
                 best_ratio=ratio
                 accgrp=ig1
              endif
           enddo

           if (best_ratio.lt.1) then
              if (good_particle) counter(7)=counter(7)+1
              if (good_particle) counter(9)=counter(9)+1
              accrflag=1
              to_group=neigh(accgrp)
              call accretion(neigh(accgrp),i_global,j_global,k_global,indices(iz),Fmax_frag(indices(iz)))
           else

! If the point has not been accreted at all:

! *********
! filament!
! *********
              if (good_particle) counter(12)=counter(12)+1
              mass(1)=mass(1)+1
              group_ID(indices(iz))=1
              link_list(indices(iz))=indices(iz)

           endif
! end of accretion if
        endif

     else
! **********************************************************************
!                       FOURTH CASE: FILAMENTS
! **********************************************************************
        mass(1)=mass(1)+1
        group_ID(indices(iz))=1
        link_list(indices(iz))=indices(iz)

! end of cases

     endif

! Checks whether to accrete all the neighbouring filaments

     if (accrflag.eq.1) then
        do ifil=1,nf
           indx=fil_list(1,ifil)+(fil_list(2,ifil)-1)*Lgrid_frag(1)+ &
                (fil_list(3,ifil)-1)*Lgrid12
!           if (fil_list(1,ifil)>safe .and. fil_list(1,ifil)<=Lgrid_frag(1)-safe .and. &
!               fil_list(2,ifil)>safe .and. fil_list(2,ifil)<=Lgrid_frag(2)-safe .and. &
!               fil_list(3,ifil)>safe .and. fil_list(3,ifil)<=Lgrid_frag(3)-safe) counter(13)=counter(13)+1
           call accretion(to_group,fil_list(1,ifil)+nstart(1),fil_list(2,ifil)+nstart(2) &
                ,fil_list(3,ifil)+nstart(3),indx,Fmax_frag(indices(iz)))
           mass(1)=mass(1)-1
        enddo
     endif


#ifdef PLC
! checks again whether groups near the PLC have passed through after the 
     if(Fmax_frag(indices(iz))<Flimit_plc) then
        do ig1=1,neigrp

           if (good(neigh(ig1)) .and. mass(neigh(ig1))>=MIN_MASS_FOR_CAT .and. check_again(1,ig1)<999) then

              replicate=check_again(:,ig1)
              thisgroup=neigh(ig1)
              aa=condition_PLC(Fmax_frag(indices(iz)))

              if (aa>=0.) then 
                 call store_plc(Fmax_frag(indices(iz)))
              endif

           endif
        enddo
     endif
#endif


! ************************************************************************
! END OF DO-CYCLE ON COLLAPSED POINTS
! ************************************************************************

  enddo

#ifdef PLC
! Past light cone construction:
! last check on halos that have not been touched recently

  if (Flimit_plc>0.) then

     replicate=0
     !dist_at_F=proper_dist(inverse_grow(1./fout(nout)))/InterPartDist

     do i=2,ngroups

        if (good(i) .and. point(i) > 1 .and. mass(i) >= MIN_MASS_FOR_CAT) then

           thisgroup=i
           bb=condition_PLC(fout(nout))

           if (bb>=0.) then
              aa=condition_PLC(Flast(i))

              if (aa<0.0) then
                 Fplc=zbrent(condition_PLC,Flast(i),fout(nout),1.d-6)
                 call store_plc(Fplc)
              endif
           endif

        end if

     end do

  endif
#endif


! Counters

99 call MPI_REDUCE(counter, all_counters, 14, MPI_INTEGER, MPI_SUM, 0, MPI_COMM_WORLD, ierr)

  if (proc_id==0) then
     write(*,'("Total number of peaks:                ",i12)')      all_counters( 0)
     write(*,'("Particles with N neighbouring groups:" ,6(1x,i12))')all_counters(1:6)
     write(*,'("Total number of accretion events:     ",i12)')      all_counters( 7)
     write(*,'("Accretion before evaluating merger:   ",i12)')      all_counters( 8)
     write(*,'("Accretion after evaluating merger:    ",i12)')      all_counters( 9)
     write(*,'("Total number of merger events:        ",i12)')      all_counters(10)
     write(*,'("Total number of major merger events:  ",i12)')      all_counters(11)
     write(*,'("Total number of filament particles:   ",i12)')      all_counters(12)
!     write(*,'("Total number of accreted filaments:   ",i12)')      all_counters(13)
  endif

! Bye!

  return
end subroutine groups


#ifdef PLC
function condition_PLC(F)

! Condition for the identification of groups in past light cone

  use pinocchio_module
  use fragment_group_vectors
  use tocompute_PLC_condition
  use tocompute_pos
  implicit none

  real (kind=doubleR)::condition_PLC
  real (kind=doubleR),intent(in)::F
  integer(kind=singleI)::i
  real (kind=doubleR)::diff1,proper_distance,z
  real (kind=doubleR),external::proper_dist,inverse_grow,q2x

!  if (F>=0.) then
  z=inverse_grow(1./F)
  proper_distance=proper_dist(z)/InterPartDist
!  else
!     proper_distance=dist_at_F
!  endif

  call set_obj(thisgroup,F,obj)

  condition_PLC=0.0
  do i=1,3
     diff1=q2x(i,obj) - ( PLCone_centre(i) + dble(Lgrid(i)*replicate(i)) )
     condition_PLC=condition_PLC+diff1**2

!     write(180+proc_id,'(f12.5,i6,f12.5)')F,thisgroup,diff1

  enddo

  condition_PLC = sqrt(condition_PLC) - proper_distance

  return
end function condition_PLC


subroutine store_plc(F)
  use pinocchio_module
  use fragment_group_vectors
  use tocompute_PLC_condition
  use tocompute_pos
  implicit none
 
  real(kind=doubleR),intent(in):: F
  integer(kind=singleI)::i,ii
  real(kind=doubleR),external::inverse_grow,q2x,vel

  NPLC=NPLC+1
  if (NPLC>=NPLC_max) then
     write(info,'("PLC count overshooted! ",i10)')NPLC_max
     call abort_code(info)
  endif

  call set_obj(thisgroup,F,obj)

  zplc(NPLC)=inverse_grow(1./F)
  Mplc(NPLC)=mass(thisgroup)*Mpart
  nameplc(NPLC)=name(thisgroup)

  do i=1,3
#ifdef ROTATE_BOX
     ii=i-1
     if (ii==0) ii=3
#else
     ii=i
#endif
     xplc(ii,NPLC)=InterPartDist * ( q2x(i,obj) - ( PLCone_centre(i) + dble(Lgrid(i)*replicate(i)) ) )
     vplc(ii,NPLC)=vel(i,obj)

  enddo

  return
end subroutine store_plc
#endif


subroutine compute_mf(iout)

! Mass function of the groups

  use pinocchio_module
  use fragment_group_vectors
  use fragment_link
  use fragment_grids

  implicit none
  integer(kind=singleI),intent(in)::iout

  integer(kind=singleI),parameter::nbin=40
  integer(kind=singleI),dimension(nbin)::ninbin,ninbin_local
  real(kind=doubleR),dimension(nbin)::massinbin,massinbin_local
  integer(kind=singleI)::i,ibin,nc
  character(len=300)::filename
  character(len=10)::label,lab1,lab2
  real(kind=doubleR)::mmin,mmax,lmass,deltam,vol,amass
  real(kind=doubleR)::x,dm,a,a1,a2,a3,Mstar,hfactor,m,D,mx
!  real(kind=doubleR),external::an_h_s
  real(kind=doubleR),external::an_h,aM_star,aLambda,Size,dLambda,grow
  integer(kind=singleI),parameter::unit=10

  character(len=24),external :: fdate

  cputmp=MPI_WTIME()

  if (OutputInH100) then
     hfactor=Hubble100
  else
     hfactor=1.0
  endif

  vol=BoxSize(1)*BoxSize(2)*BoxSize(3)
  Mstar  = aM_star(zout(iout))
  D=grow(zout(iout))

  mmin=log10(10.*Mpart) ; mmax=16.5
  deltam=(mmax-mmin)/float(nbin)
  lmass=0.
  nc=0 ; ninbin=0; ninbin_local=0; massinbin=0.0; massinbin_local=0.0

  do i=2,ngroups
     if (point(i) > 0 .and. good(i)) then
        amass=log10(mass(i)*Mpart)
        if (amass.gt.mmin) nc=nc+1
        lmass=max(amass,lmass)
        ibin=int((amass-mmin)/deltam+1.)
        if (ibin < 1 .or. ibin > nbin) cycle
        ninbin_local(ibin)=ninbin_local(ibin)+1
        massinbin_local(ibin)=massinbin_local(ibin)+mass(i)*Mpart
     endif
  enddo

  call MPI_REDUCE(ninbin_local, ninbin, nbin, MPI_INTEGER, MPI_SUM, 0, MPI_COMM_WORLD, ierr)
  call MPI_REDUCE(massinbin_local, massinbin, nbin, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD, ierr)

! Writes result

  if (proc_id==0) then
     write(label,'(f10.4)')zout(iout)
     filename='pinocchio.'//trim(adjustl(label))//'.'//trim(run)//'.mf.out'
     write(info,'("[",a24,"] Writing file ",a)')fdate(), trim(filename)
     call message(trim(info))
     open(unit,file=filename,status='unknown')
     write(unit,'("# Mass function for redshift ",f8.4)')zout(iout)
     if (OutputInH100) then 
!        write(unit,'("# All quantities are relative to H=100 km/s/Mpc")')
        lab1='/h'
        lab2='h^4'
     else
        write(unit,'("# All quantities are relative to H=",f8.4," km/s/Mpc")')Hubble100*100.
     endif
     write(unit,'("# 1) mass (Msun",a,")")')trim(lab1)
     write(unit,'("# 2) n(m) (Mpc^-3 Msun^-1 ",a,")")')trim(lab2)
     write(unit,'("# 3) upper +1-sigma limit for n(m) (Mpc^-3 Msun^-1 ",a,")")')trim(lab2)
     write(unit,'("# 4) lower -1-sigma limit for n(m) (Mpc^-3 Msun^-1 ",a,")")')trim(lab2)
     write(unit,'("# 5) number of halos in the bin")')
     select case (flag_ps)
     case(0)
        write(unit,'("# 6) analytical n(m) from Press & Schechter 1974")')
     case(1)
        write(unit,'("# 6) analytical n(m) from Sheth & Tormen 2001")')
     case(2)
        write(unit,'("# 6) analytical n(m) from Jenkins et al. 2001")')
     case(3)
        write(unit,'("# 6) analytical n(m) from Warren et al. 2006")')
     case(4)
        write(unit,'("# 6) analytical n(m) from Reed et al. 2007")')
     case(5)
        write(unit,'("# 6) analytical n(m) from Crocce et al. 2010")')
     case(6)
        write(unit,'("# 6) analytical n(m) from Tinker et al. 2010")')
     case(7)
        write(unit,'("# 6) analytical n(m) from Courtin et al. 2010")')
     end select
     write(unit,'("#")')

     do i=1,nbin
        x=mmin+(i-.5)*deltam
        m=10.**x
        dm=10.**(mmin+i*deltam)-10.**(mmin+(i-1)*deltam)
        a=ninbin(i)/vol/dm
        a1=(ninbin(i)+sqrt(float(ninbin(i))))/vol/dm
        a2=(ninbin(i)-sqrt(float(ninbin(i))))/vol/dm
        a3=an_h(m,zout(iout))
        if (ninbin(i) > 5) then 
           mx=massinbin(i)/dble(ninbin(i))
        else 
           mx=m
        endif

        write(unit,'(4(1x,g15.8),3x,i10,3x,g15.8,3x,3(1x,g15.8))') &
             mx*hfactor,      &
             a/hfactor**4,    &
             a1/hfactor**4,   &
             a2/hfactor**4,   &
             ninbin(i),       &
             a3/hfactor**4
     enddo
     close(unit)
  endif
! Bye!

  cputime(5)=cputime(5)+MPI_WTIME()-cputmp

  return
end subroutine compute_mf


subroutine merge_groups(grp1,grp2,time)

! Merges grp2 into grp1

  use pinocchio_module
  use fragment_grids
  use fragment_link
  use fragment_merger
  use fragment_group_vectors
  use tocompute_pos
  implicit none

  integer(kind=singleI),intent(in)::grp1,grp2
  real(kind=doubleR),intent(in)::time

  integer(kind=singleI)::i1

! MAJOR MERGING EVENTS

  IF (mass(GRP2).gt.mass(GRP1)/2..AND.MASS(GRP2).GT.MASSMIN) THEN
     MAJORZ(GRP1)=TIME
  ENDIF

! updates the linking list

  i1=point(grp2)
  do while (link_list(i1).ne.point(grp2))
     group_ID(i1)=grp1
     i1=link_list(i1)
  enddo
  group_ID(i1)=grp1

  link_list(bottom(grp1))=point(grp2)
  link_list(bottom(grp2))=point(grp1)
  bottom (grp1)=bottom(grp2)
  point(grp2)=-1
  bottom (grp2)=-1

! UPDATES THE SUBSTRUCTURE IN THE NEW GROUP

  if (mass(grp1).gt.MASSMIN.and.mass(grp2).gt.MASSMIN) then
     CALL UPDATESUBSTRUCT(GRP1,GRP2,TIME)
  endif

! updates the centre position and masses, and the existence flag

  call set_obj(grp1,time,obj1)
  call set_obj(grp2,time,obj2)

  call update()

  call set_group(grp1,obj1)
  
  if (mass(grp1).gt.MASSMIN.and.T_APPEAR(grp1).eq.-1) &
       T_APPEAR(grp1)=time

! Bye!

  return
end subroutine merge_groups



SUBROUTINE UPDATESUBSTRUCT(G1,G2,TIME)

! UPGRADE THE SUBSTRUCTURE OF A GROUP: M(G1)>M(G2) IE G2 --> G1 

  use pinocchio_module
  use fragment_merger
  use fragment_group_vectors
  use fragment_grids

  IMPLICIT NONE
  integer(kind=singleI),intent(in)::G1,G2
  real(kind=doubleR),intent(in)::time

  integer(kind=singleI)::old_i,first_g2


! IF GROUPS HAVE NO SUBSTRUCTURE
  IF(GR_LL(G1).EQ.G1.AND.GR_LL(G2).EQ.G2) THEN 
!     MCENTR(G1)=MASS(G1)
!     MCENTR(G2)=MASS(G2)
     GR_LL(G1)=G2
     GR_LL(G2)=G1

! IF G1 HAS SUBSTRUCTURE BUT G2 IS A SINGLEHALO
  ELSEIF(GR_LL(G1).NE.G1.AND.GR_LL(G2).EQ.G2) THEN
! old procedure
!     GR_LL(G2)=GR_LL(G1)
!     GR_LL(G1)=G2
! new procedure
     gr_ll(g2)=g1
     old_i=g1
     do while (gr_ll(old_i).ne.g1)
        old_i=gr_ll(old_i)
     enddo
     gr_ll(old_i)=g2
!     end new procedure
!     MCENTR(G2)=MASS(G2)

! IF G1 IS A SINGLE HALO AND G2 HAS SUBSTRUCTURE (RARE!)
  ELSEIF(GR_LL(G1).EQ.G1.AND.GR_LL(G2).NE.G2) THEN
     OLD_I=G2
     DO WHILE (GR_LL(OLD_I).NE.G2)
        OLD_I=GR_LL(OLD_I)
        HALO_APP(OLD_I)=G1
     ENDDO
     HALO_APP(G2)=G1
     GR_LL(G1)=GR_LL(G2)
     GR_LL(G2)=G1
!     MCENTR(G1)=MASS(G1)
! BOTH THE GROUPS HAVE SUBSTRUCTURE
  ELSEIF(GR_LL(G1).NE.G1.AND.GR_LL(G2).NE.G2) THEN
     OLD_I=G2
     DO WHILE (GR_LL(OLD_I).NE.G2)
        OLD_I=GR_LL(OLD_I)
        HALO_APP(OLD_I)=G1
     ENDDO
! old procedure
!         INDX=GR_LL(G1)
!         GR_LL(G1)=GR_LL(G2)
!         GR_LL(G2)=INDX
! new procedure
     first_g2=old_i
     old_i=g1
     do while (gr_ll(old_i).ne.g1)
        old_i=gr_ll(old_i)
     enddo
     gr_ll(old_i)=gr_ll(g2)
     gr_ll(g2)=g1
! end new procedure

  ENDIF

  HALO_APP(G2)=G1
  T_ACC(G2)=TIME
  MASS_AT_MERGER(G2)=MASS(G1)
  MERGED_WITH(G2)=G1

  RETURN
END SUBROUTINE UPDATESUBSTRUCT



subroutine accretion(group,i,j,k,indx,F)

! Accretes the point (i,j,k) on the group

  use pinocchio_module
  use fragment_merger
  use fragment_group_vectors
  use fragment_grids
  use fragment_matrices
  use fragment_link
  use tocompute_pos
  implicit none

  integer(kind=singleI),intent(in)::group,i,j,k,indx
  real(kind=doubleR),intent(in)::F

  call set_obj(group,F,obj1)
  call set_point(i,j,k,indx,F,obj2)

  call update()

  call set_group(group,obj1)

  if (mass(group).gt.MASSMIN.and.T_APPEAR(group).eq.-1) T_APPEAR(group)=F
  group_ID(indx)=group

! Updates the linking list

  link_list(bottom(group))=indx
  bottom(group)=indx
  link_list(indx)=point(group)

!  write(120+proc_id,'(f12.5,4(1x,i6),9(1x,f12.5))')F,group,i,j,k,mass(group),centres(:,group),velgrp(:,group)

  return
end subroutine accretion


subroutine write_cat(iout)

! Writes the group catalogues

  use pinocchio_module
  use fragment_run_flags
  use fragment_grids
  use fragment_link
  use fragment_group_vectors
  use fragment_merger
  use tocompute_pos
  implicit none

  integer(kind=singleI),intent(in)::iout
  integer(kind=singleI),parameter::unit=10
  integer(kind=singleI)::igood,i,ngood,nhalos,j,iproc

  integer(kind=doubleI),allocatable,dimension(:,:)::ivector
  real   (kind=doubleR),allocatable,dimension(:,:)::rvector
  integer(kind=singleI)::nrdata,nidata

  real(kind=doubleR)::vf,hfactor
  character(len=100)::filename
  character(len=10)::label
  character(len=2)::labh
  real(kind=doubleR),external::inverse_grow,q2x,vel

! ordering of coordinates to accomodate for rotation caused by fft ordering
#ifdef ROTATE_BOX
  integer(kind=singleI),parameter:: x1=2, x2=3, x3=1
#else
  integer(kind=singleI),parameter:: x1=1, x2=2, x3=3
#endif
#ifdef WRITE_GROUPIDS
  integer(kind=singleI)::Lgrid12,iz,k,kk,i_global,j_global,k_global
  logical::good_particle
#endif


  character(len=24),external :: fdate

  cputmp=MPI_WTIME()

! output in H100 or in Htrue
  if (OutputInH100) then
     hfactor=Hubble100
  else
     hfactor=1.0
  endif

! ******* !
! CATALOG !
! ******* !

  nrdata=9
  nidata=2

! each processor builds the catalogue
  ngood=0
  nhalos=0
  do i=2,ngroups
     if (point(i) > 0 .and. good(i) .and. mass(i) >= MIN_MASS_FOR_CAT) ngood=ngood+1
  enddo

  if (ngood>0) then
     allocate(ivector(0:nidata-1,ngood))
     allocate(rvector(0:nrdata-1,ngood))

     igood=1
     do i=2,ngroups
        if (point(i) > 0 .and. good(i) .and. mass(i) >= MIN_MASS_FOR_CAT) then

           call set_obj(i,fout(iout),obj)

           ivector(0,igood)=name(i)
           ivector(1,igood)=mass(i)

           rvector(0,igood)=centres(x1,i)*InterPartDist*hfactor
           rvector(1,igood)=centres(x2,i)*InterPartDist*hfactor
           rvector(2,igood)=centres(x3,i)*InterPartDist*hfactor
           rvector(3,igood)=q2x(x1,obj)*InterPartDist*hfactor
           rvector(4,igood)=q2x(x2,obj)*InterPartDist*hfactor
           rvector(5,igood)=q2x(x3,obj)*InterPartDist*hfactor
           rvector(6,igood)=vel(x1,obj)
           rvector(7,igood)=vel(x2,obj)
           rvector(8,igood)=vel(x3,obj)

           ! PBCs on displaced groups
           do j=1,3
              if (rvector(2+j,igood)<0) rvector(2+j,igood)=rvector(2+j,igood)+BoxSize(j)*hfactor
              if (rvector(2+j,igood)>=BoxSize(j)*hfactor) rvector(2+j,igood)=rvector(2+j,igood)-BoxSize(j)*hfactor
           enddo

           igood=igood+1
        endif
     enddo
  endif

! Processor 0 opens the file and writes its catalogue
  if (proc_id==0) then

     write(label,'(f10.4)')zout(iout)
     filename='pinocchio.'//trim(adjustl(label))//'.'//trim(run)//'.catalog.out'

     write(info,'("[",a24,"] Writing file ",a)')fdate(), filename
     call message(trim(info))

     if (CatalogInAscii) then
        open(unit,file=trim(filename),status='unknown',form='formatted')
        write(unit,'("# Group catalog for redshift ",f8.4," and minimal mass of ",i6," particles")')zout(iout), MIN_MASS_FOR_CAT
        if (OutputInH100) then
!           write(unit,'("# All quantities are relative to H0=100 km/s/Mpc")')
           labh='/h'
        else
!           write(unit,'("# All quantities are relative to H0=",f6.3," km/s/Mpc")')Hubble100*100.0
           labh=''
        endif
        write(unit,'("#    1) group ID")')
        write(unit,'("#    2) group mass (Msun",a,")")')trim(labh)
        write(unit,'("# 3- 5) initial position (Mpc",a,")")')trim(labh)
        write(unit,'("# 6- 8) final position (Mpc",a,")")')trim(labh)
        write(unit,'("# 9-11) velocity (km/s)")')
        write(unit,'("#   12) number of particles")')
        write(unit,'("#")')

        if (ngood>0) then
           do igood=1,ngood
              write(unit,'(1x,i12,1x,e13.6,9(1x,f10.2),1x,i12)') &
                   ivector(0,igood),ivector(1,igood)*Mpart*hfactor,rvector(0:nrdata-1,igood),ivector(1,igood)
           enddo
! and then deallocates its vector
           deallocate (ivector)
           deallocate (rvector)
        endif
        nhalos=nhalos+ngood

     else

        open(unit,file=trim(filename),status='unknown',form='unformatted')

        write(unit)nproc
        write(unit)ngood

        if (ngood>0) then
           do igood=1,ngood
              write(unit) &
                   ivector(0,igood),ivector(1,igood)*Mpart*hfactor,rvector(0:nrdata-1,igood),int(ivector(1,igood))
           enddo
! and then deallocates its vector
           deallocate (ivector)
           deallocate (rvector)
        endif
        nhalos=nhalos+ngood

     endif

  endif

! loop on other processors
  do iproc=1,nproc-1

! processor i sends the number of good groups, processor 0 receives it
     if (proc_id==0) then
        ngood=0
        call MPI_RECV(ngood, 1, MPI_INTEGER, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)
! then processor i sends the groups to processor 0 and deallocates the vectors, 
! while processor 0 allocates the vectors and receives the groups
        if (ngood>0) then
           allocate(ivector(0:nidata-1,ngood))
           allocate(rvector(0:nrdata-1,ngood))
           call MPI_RECV(ivector, nidata*ngood, MPI_DOUBLE_PRECISION, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)
           call MPI_RECV(rvector, nrdata*ngood, MPI_DOUBLE_PRECISION, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)
        endif
     else if (proc_id==iproc) then
        call MPI_SEND(ngood, 1, MPI_INTEGER, 0, 0, MPI_COMM_WORLD, ierr)
        if (ngood>0) then
           call MPI_SEND(ivector, nidata*ngood, MPI_DOUBLE_PRECISION, 0, 0, MPI_COMM_WORLD, ierr)
           call MPI_SEND(rvector, nrdata*ngood, MPI_DOUBLE_PRECISION, 0, 0, MPI_COMM_WORLD, ierr)
           deallocate(ivector)
           deallocate(rvector)
        endif
     endif

! processor 0 writes the groups
     if (proc_id==0) then
        if (ngood>0) then

           if (CatalogInAscii) then
              do igood=1,ngood
                 write(unit,'(1x,i12,1x,e13.6,9(1x,f10.2),1x,i12)') &
                      ivector(0,igood),ivector(1,igood)*Mpart*hfactor,rvector(0:nrdata-1,igood),ivector(1,igood)
              enddo
           else
              write(unit)ngood
              do igood=1,ngood
                 write(unit) &
                      ivector(0,igood),ivector(1,igood)*Mpart*hfactor,rvector(0:nrdata-1,igood),int(ivector(1,igood))
              enddo
           endif

           deallocate(ivector)
           deallocate(rvector)
        endif
        nhalos=nhalos+ngood
     endif

! end of loop on processors
  enddo

  if (proc_id==0) close(unit)

  write(info,'("[",a24,"] ",i9," halos written on file ",a)')fdate(), nhalos, filename
  call message(trim(info))


#ifdef WRITE_GROUPIDS

! ********* !
! GROUP IDs !
! ********* !

! This is a rather lengthy output, you may not want to have it in standard runs
! Also, it is convenient to output also the safety boundaries, because you don't want
! to waste time here in deciding what to do in the overlapping regions.
! Ergo, each task writes its own file, and you switch this option on only when you really need it.

  write(label,'(f10.4)')zout(iout)
  filename='pinocchio.'//trim(adjustl(label))//'.'//trim(run)//'.groupIDs.out'
  write(label,'(i10)')proc_id
  filename=trim(filename)//'.'//trim(adjustl(label))

  open(unit,file=filename,form='unformatted',status='unknown')
  write(unit)Npart_frag
  Lgrid12=Lgrid_frag(1)*Lgrid_frag(2)
  do iz=1,Npart_frag

     k=1+(iz-1)/Lgrid12
     kk=iz-1-(k-1)*Lgrid12
     j=1+kk/Lgrid_frag(1)
     i=1+(kk-(j-1)*Lgrid_frag(1))

     i_global=i+nstart(1)
     j_global=j+nstart(2)
     k_global=k+nstart(3)

     good_particle=(i>safe .and. i<=Lgrid_frag(1)-safe .and. j>safe .and. j<=Lgrid_frag(2)-safe .and. k>safe .and. k<=Lgrid_frag(3)-safe)

     if (group_ID(iz)==0) then
        write(unit)good_particle,i_global,j_global,k_global,int(-99,8)
     else if (group_ID(iz)==1) then
        write(unit)good_particle,i_global,j_global,k_global,int(-1,8)
     else
        write(unit)good_particle,i_global,j_global,k_global,name(group_ID(iz))
     endif

  enddo
  close(unit)

#endif

! merger histories are written only at the final time

  if (iout /= nout) then
     cputime(5)=cputime(5)+MPI_WTIME()-cputmp
     return
  endif

! **************** !
! MERGER HISTORIES !
! **************** !

  nrdata=3
  nidata=6

! each processor builds the catalogue

  ngood=0
  nhalos=0

  do i=2,ngroups
     if (halo_app(i)>1) then
        if (good(halo_app(i)) .and. mass(halo_app(i)) >= MIN_MASS_FOR_CAT) ngood=ngood+1
     endif
  enddo

  if (ngood>0) then
     allocate(ivector(0:nidata-1,ngood))
     allocate(rvector(0:nrdata-1,ngood))
  
     igood=1
     do i=2,ngroups

        if (good(halo_app(i)) .and. mass(halo_app(i)) >= MIN_MASS_FOR_CAT) then
           ivector(0,igood)=name(i)
           ivector(1,igood)=name(halo_app(i))
           ivector(2,igood)=name(gr_ll(i))
           if (merged_with(i)>1) then
              ivector(3,igood)=name(merged_with(i))
           else
              ivector(3,igood)=-1
           endif
           ivector(4,igood)=mass(i)
           ivector(5,igood)=mass_at_merger(i)

           if (t_acc(i)>0.) then
              rvector(0,igood)=inverse_grow(1./t_acc(i))
           else
              rvector(0,igood)=t_acc(i)
           endif
           if (majorz(i)>0.) then
              rvector(1,igood)=inverse_grow(1./majorz(i))
           else
              rvector(1,igood)=majorz(i)
           endif
           if (t_appear(i)>0.) then
              rvector(2,igood)=inverse_grow(1./t_appear(i))
           else
              rvector(2,igood)=t_appear(i)
           endif

           igood=igood+1
        endif
     enddo
  endif


! Processor 0 opens the file and writes its catalogue
  if (proc_id==0) then

     filename='pinocchio.'//trim(run)//'.histories.out'

     write(info,'("[",a24,"] Writing file ",a)')fdate(), filename
     call message(trim(info))

     if (CatalogInAscii) then

        open(unit,file=trim(filename),status='unknown',form='formatted')
        write(unit,'("# Merger histories for halos with minimal mass of ",i5," particles")') MIN_MASS_FOR_CAT
        write(unit,'("#  1) group ID")')
        write(unit,'("#  2) belongs to")')
        write(unit,'("#  3) linking list")')
        write(unit,'("#  4) merged with")')
        write(unit,'("#  5) mass of halo at merger (particles)")')
        write(unit,'("#  6) mass of main halo it merges with, at merger (particles)")')
        write(unit,'("#  7) unused")')
        write(unit,'("#  8) merger redshift")')
        write(unit,'("#  9) redshift of last major merger ")')
        write(unit,'("# 10) appearance redshift")')
        write(unit,'("#")')

        if (ngood>0) then
           do i=1,ngood
              write(unit,'(4(1x,i12),3(1x,i9),3(1x,f9.4))') &
                   ivector(0:3,i),int(ivector(4,i)),int(ivector(5,i)),0, &
                   rvector(:,i)
           enddo
! and then deallocates its vector
           deallocate (ivector)
           deallocate (rvector)
        endif
        nhalos=nhalos+ngood

     else

        open(unit,file=trim(filename),status='unknown',form='unformatted')

        write(unit)nproc
        write(unit)ngood

        if (ngood>0) then
           do i=1,ngood
              write(unit)ivector(0:3,i),int(ivector(4,i)),int(ivector(5,i)), &
                   rvector(:,i)
           enddo
! and then deallocates its vector
           deallocate (ivector)
           deallocate (rvector)
        endif
        nhalos=nhalos+ngood

     endif

  endif

! loop on other processors
  do iproc=1,nproc-1

! processor i sends the number of good groups, processor 0 receives it
     if (proc_id==0) then

        ngood=0
        call MPI_RECV(ngood, 1, MPI_INTEGER, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)

        if (ngood>0) then
           allocate(ivector(0:nidata-1,ngood))
           allocate(rvector(0:nrdata-1,ngood))
           call MPI_RECV(ivector, nidata*ngood, MPI_DOUBLE_PRECISION, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)
           call MPI_RECV(rvector, nrdata*ngood, MPI_DOUBLE_PRECISION, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)
        endif

! then processor i sends the groups to processor 0 and deallocates the vectors, 
! while processor 0 allocates the vectors and receives the groups

     else if (proc_id==iproc) then

        call MPI_SEND(ngood, 1, MPI_INTEGER, 0, 0, MPI_COMM_WORLD, ierr)

        if (ngood>0) then
           call MPI_SEND(ivector, nidata*ngood, MPI_DOUBLE_PRECISION, 0, 0, MPI_COMM_WORLD, ierr)
           call MPI_SEND(rvector, nrdata*ngood, MPI_DOUBLE_PRECISION, 0, 0, MPI_COMM_WORLD, ierr)
           deallocate(ivector)
           deallocate(rvector)
        endif

     endif

! processor 0 writes the groups
     if (proc_id==0) then

        if (CatalogInAscii) then
           do i=1,ngood
              write(unit,'(4(1x,i12),3(1x,i9),3(1x,f9.4))') &
                   ivector(0:3,i),int(ivector(4,i)),int(ivector(5,i)),0, &
                   rvector(:,i)
           enddo
        else
           write(unit)ngood
           do i=1,ngood
              write(unit)ivector(0:3,i),int(ivector(4,i)),int(ivector(5,i)), &
                   rvector(:,i)
           enddo
        endif

        if (ngood>0) then
           deallocate(ivector)
           deallocate(rvector)
        endif

     endif

     nhalos=nhalos+ngood

! end of loop on processors
  enddo

  if (proc_id==0) close(unit)

  write(info,'("[",a24,"] merger histories written on file ",a)')fdate(), filename
  call message(trim(info))

  cputime(5)=cputime(5)+MPI_WTIME()-cputmp
  return
end subroutine write_cat

#ifdef PLC
subroutine write_plc()

! Writes the groups in the PLC

  use pinocchio_module
  use fragment_group_vectors
  use tocompute_PLC_condition
  implicit none

  integer(kind=singleI),parameter::unit=10
  character(len=100)::filename
  character(len=2)::labh
  character(len=24),external :: fdate
  integer(kind=singleI)::nhalos,i,iproc
  real(kind=doubleR)::hfactor,rhor,theta,phi,v_los,obsz
  integer(kind=singleI)::nbinz
  integer(kind=singleI),allocatable::histogram

  cputmp=MPI_WTIME()

! output in H100 or in Htrue
  if (OutputInH100) then
     hfactor=Hubble100
  else
     hfactor=1.0
  endif

! Processor 0 opens the file and writes its catalogue
  if (proc_id==0) then

     filename='pinocchio.'//trim(run)//'.plc.out'

     write(*,'("[",a24,"] Writing file ",a)')fdate(), filename

     if (CatalogInAscii) then
        open(unit,file=trim(filename),status='unknown',form='formatted')
        write(unit,'("# Group catalog on the Past Light Cone for a minimal mass of ",i5," particles")') MIN_MASS_FOR_CAT
        if (OutputInH100) then
!           write(unit,'("# All quantities are relative to H0=100 km/s/Mpc")')
           labh='/h'
        else
!           write(unit,'("# All quantities are relative to H0=",f6.3," km/s/Mpc")')Hubble100*100.0
           labh=''
        endif
        write(unit,'("#    1) group ID")')
        write(unit,'("#    2) true redshift")')
        write(unit,'("#  3-5) comoving position (Mpc",a,")")')trim(labh)
        write(unit,'("#  6-8) velocity (km/s)")')
        write(unit,'("#    9) group mass (Msun",a,")")')trim(labh)
        write(unit,'("#   10) theta (degree)")')
        write(unit,'("#   11) phi (degree)")')
        write(unit,'("#   12) peculiar velocity along the line-of-sight (km/s)")')
        write(unit,'("#   13) observed redshift")')
        write(unit,'("#")')

        if (NPLC>0) then
           do i=1,NPLC
              call coor_trasformation_cartesian_polar(xplc(:,i),rhor,theta,phi)
              v_los=(xplc(1,i)*vplc(1,i)+xplc(2,i)*vplc(2,i)+xplc(3,i)*vplc(3,i))/rhor
              obsz=zplc(i)+v_los/299792.458
              write(unit,'(1x,i12,7(1x,f16.6),1x,e15.8,4(1x,f16.6))') &
                   nameplc(i), &
                   zplc(i),    &
                   xplc(:,i)*hfactor,  &
                   vplc(:,i),  &
                   Mplc(i)*hfactor,    &
                   theta,      &
                   phi,        &
                   v_los,      &
                   obsz
              
           enddo
        endif

     else

        open(unit,file=trim(filename),status='unknown',form='unformatted')

        if (NPLC>0) then
           do i=1,NPLC
              call coor_trasformation_cartesian_polar(xplc(:,i),rhor,theta,phi)
              v_los=(xplc(1,i)*vplc(1,i)+xplc(2,i)*vplc(2,i)+xplc(3,i)*vplc(3,i))/rhor
              obsz=zplc(i)+v_los/299792.458
              write(unit)      &
                   nameplc(i), &
                   zplc(i),    &
                   xplc(:,i)*hfactor,  &
                   vplc(:,i),  &
                   mplc(i)*hfactor,    &
                   theta,      &
                   phi,        &
                   v_los,      &
                   obsz
           enddo
        endif

     endif

     nhalos=nhalos+NPLC

     deallocate(zplc)
     deallocate(Mplc)
     deallocate(nameplc)
     deallocate(xplc)
     deallocate(vplc)

  endif

! loop on other processors
  do iproc=1,nproc-1

! processor i sends the number of groups in the PLC, processor 0 receives it
     if (proc_id==0) then
        NPLC=0
        call MPI_RECV(NPLC, 1, MPI_INTEGER, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)

! then processor i sends the groups to processor 0 and deallocates the vectors, 
! while processor 0 allocates the vectors and receives the groups

        if (NPLC>0) then

           allocate (zplc   (NPLC))
           allocate (Mplc   (NPLC))
           allocate (xplc (3,NPLC))
           allocate (vplc (3,NPLC))
           allocate (nameplc(NPLC))

           call MPI_RECV(zplc,    NPLC, MPI_DOUBLE_PRECISION, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)
           call MPI_RECV(Mplc,    NPLC, MPI_DOUBLE_PRECISION, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)
           call MPI_RECV(xplc,  3*NPLC, MPI_DOUBLE_PRECISION, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)
           call MPI_RECV(vplc,  3*NPLC, MPI_DOUBLE_PRECISION, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)
           call MPI_RECV(nameplc, NPLC, MPI_DOUBLE_PRECISION, iproc, 0, MPI_COMM_WORLD, MPI_STATUS, ierr)

        endif

     else if (proc_id==iproc) then
        call MPI_SEND(NPLC, 1, MPI_INTEGER, 0, 0, MPI_COMM_WORLD, ierr)

        if (NPLC>0) then

           call MPI_SEND(zplc,    NPLC, MPI_DOUBLE_PRECISION, 0, 0, MPI_COMM_WORLD, ierr)
           call MPI_SEND(Mplc,    NPLC, MPI_DOUBLE_PRECISION, 0, 0, MPI_COMM_WORLD, ierr)
           call MPI_SEND(xplc,  3*NPLC, MPI_DOUBLE_PRECISION, 0, 0, MPI_COMM_WORLD, ierr)
           call MPI_SEND(vplc,  3*NPLC, MPI_DOUBLE_PRECISION, 0, 0, MPI_COMM_WORLD, ierr)
           call MPI_SEND(nameplc, NPLC, MPI_DOUBLE_PRECISION, 0, 0, MPI_COMM_WORLD, ierr)

           deallocate(zplc)
           deallocate(Mplc)
           deallocate(nameplc)
           deallocate(xplc)
           deallocate(vplc)

        endif

     endif

! processor 0 writes the groups
     if (proc_id==0) then

        if (NPLC>0) then
           if (CatalogInAscii) then

              do i=1,NPLC
                 call coor_trasformation_cartesian_polar(xplc(:,i),rhor,theta,phi)
                 v_los=(xplc(1,i)*vplc(1,i)+xplc(2,i)*vplc(2,i)+xplc(3,i)*vplc(3,i))/rhor
                 obsz=zplc(i)+v_los/299792.458
                 write(unit,'(1x,i12,7(1x,f16.6),1x,e15.8,4(1x,f16.6))') &
                      nameplc(i), &
                      zplc(i),    &
                      xplc(:,i)*hfactor,  &
                      vplc(:,i),  &
                      Mplc(i)*hfactor,    &
                      theta,      &
                      phi,        &
                      v_los,      &
                      obsz
              enddo

           else

              do i=1,NPLC
                 call coor_trasformation_cartesian_polar(xplc(:,i),rhor,theta,phi)
                 v_los=(xplc(1,i)*vplc(1,i)+xplc(2,i)*vplc(2,i)+xplc(3,i)*vplc(3,i))/rhor
                 obsz=zplc(i)+v_los/299792.458
                 write(unit) &
                      nameplc(i), &
                      zplc(i),    &
                      xplc(:,i)*hfactor,  &
                      vplc(:,i),  &
                      Mplc(i)*hfactor,    &
                      theta,      &
                      phi,        &
                      v_los,      &
                      obsz
              enddo

           endif

           deallocate(zplc)
           deallocate(Mplc)
           deallocate(nameplc)
           deallocate(xplc)
           deallocate(vplc)

        endif

        nhalos=nhalos+NPLC
     endif

! end of loop on processors
  enddo

  if (proc_id==0) close(unit)

  write(info,'("[",a24,"] ",i9," halos written on file ",a)')fdate(), nhalos, filename
  call message(trim(info))

  cputime(5)=cputime(5)+MPI_WTIME()-cputmp
  return
end subroutine write_plc
#endif

SUBROUTINE clean_list(n,arr)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::n
  integer(kind=singleI),intent(inout),dimension(n)::arr

  INTEGER(KIND=SINGLEI)::i,j,a
  logical::done

  do j=2,n

     do i=j-1,1,-1
        if (arr(i).eq.arr(j)) arr(j)=0
     enddo

     a=arr(j)
     done=.false.
     do i=j-1,1,-1
        if (arr(i)>0)  then
           done=.true.
           exit
        endif
        arr(i+1)=arr(i)
     enddo
     if (.not.done) i=0
     arr(i+1)=a
  enddo
  return
END SUBROUTINE clean_list



real(kind=doubleR) function virial(grp,F,flag)

! Gives the virial radius

  use fragment_virial_params
  use pinocchio_module
  implicit none

  integer(kind=singleI),intent(in)::grp,flag
  real(kind=doubleR),intent(in)::F

  real(kind=doubleR)::rlag,zero,corr,variance
  real(kind=doubleR),external::aLambda

  rlag=float(grp)**.3333333333333333333333
  if (flag.eq.0) then
! merging
     virial=rlag*f_m + f_rm
  else
! accretion
     variance=aLambda(rlag*InterPartDist)/F**2
     corr=(variance)**(0.5*espo)
!     corr=min(corr,5.d0)
     zero=f_ra+f_s*rlag*corr
     virial=f_a*rlag+zero

!     if (proc_id==0) print '(1x,i10,99(1x,f12.6))',grp, rlag, variance, corr, zero, virial

  endif

  return
end function virial


#ifdef PLC
SUBROUTINE clean_list_with_PLC(n,arr,found)

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::n
  integer(kind=singleI),intent(inout),dimension(3,n)::found
  integer(kind=singleI),intent(inout),dimension(n)::arr

  INTEGER(KIND=SINGLEI)::i,j,a,b(3)
  logical::done

  do j=2,n

     do i=j-1,1,-1
        if (arr(i).eq.arr(j)) then
           arr(j)=0
           found(:,j)=0
        endif
     enddo

     a=arr(j)
     b=found(:,j)
     done=.false.
     do i=j-1,1,-1
        if (arr(i)>0)  then
           done=.true.
           exit
        endif
        arr(i+1)=arr(i)
        found(:,i+1)=found(:,i)
     enddo
     if (.not.done) i=0
     arr(i+1)=a
     found(:,i+1)=b
  enddo
  return

END SUBROUTINE clean_list_with_PLC
#endif


subroutine sort(n,vector)

! very simple sorting subroutine

  use numbers
  implicit none
  integer(kind=singleI),intent(in)::n
  real(kind=doubleR),intent(inout)::vector(n)

! local variables
  integer(kind=singleI)::i,j
  real(kind=doubleR)::a

  c2: do j=2,n
     a=vector(j)
     c1: do i=j-1,1,-1
        if (vector(i)>a) then
           vector(i+1)=a
           cycle c2
        else
           vector(i+1)=vector(i)
        endif
     enddo c1
     vector(i+1)=a
  enddo c2

  return
end subroutine sort


subroutine fragment_redistribute_big_matrices()

  use fragment_grids
  use fragment_link
  use fragment_matrices
  use fmax_grids
  use pinocchio_module
  implicit none
  integer(kind=singleI)::err
  character(len=24), external :: fdate

! Allocations of Fmax_frag and velocities
  call allocate_fragment_vectors()

  if (alloc_verbose) then
     write(info,"('Total allocated memory:',f8.3,' Gbyte; max allocated memory: ',f8.3,' Gbyte')") &
          total_allocated/Gbyte,max_allocated/Gbyte
     call message (trim(info))
  endif

! syncronize processors before distributing boxes
  call MPI_BARRIER(MPI_COMM_WORLD, ierr)

! distribution of Fmax_frag and velocity planes into sub-boxes
  call distribute()

! deallocates Fmax and vmax distributed in planes
  call deallocate_all_large_matrices()

  write(info,"('Total allocated memory:',f8.3,' Gbyte; max allocated memory: ',f8.3,' Gbyte')") &
       total_allocated/Gbyte,max_allocated/Gbyte
  call message (trim(info))

! other allocations

  total_allocated = total_allocated + sizeof_indices*Npart_frag
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating indices...')
  allocate (indices  (Npart_frag),stat=err)
  if (err>0) then
     call abort_code('error in redistributiongrep : unable to allocate indices')
  endif
  if (alloc_verbose) call message ('indices allocated')


  total_allocated = total_allocated + sizeof_groupID*Npart_frag
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating group_ID...')
  allocate (group_ID (Npart_frag),stat=err)
  if (err>0) then
     call abort_code('error in redistribution: unable to allocate group_ID')
  endif
  if (alloc_verbose) call message ('group_ID allocated')

  total_allocated = total_allocated + sizeof_linklist*Npart_frag
  max_allocated=max(total_allocated,max_allocated)
  if (alloc_verbose) call message ('allocating link_list...')
  allocate (link_list(Npart_frag),stat=err)
  if (err>0) then
     call abort_code('error in redistribution: unable to allocate link_list')
  endif
  if (alloc_verbose) call message ('link_list allocated')

  return
end subroutine fragment_redistribute_big_matrices


subroutine fragment_allocate_group_vectors()

  use pinocchio_module
  use fragment_group_vectors
  use fragment_merger
  use fragment_grids
  use fragment_link
  implicit none
  integer(kind=singleI)::err

  total_allocated = total_allocated + sizeof_groups*ngroups_tot
  max_allocated=max(total_allocated,max_allocated)

  allocate (mass   (  ngroups_tot),stat=err)
  allocate (centres(3,ngroups_tot),stat=err)
  allocate (velgrp (3,ngroups_tot),stat=err)
  allocate (gr_ll         (ngroups_tot),stat=err)
!  allocate (Mcentr        (ngroups_tot),stat=err)
  allocate (halo_app      (ngroups_tot),stat=err)
  allocate (mass_at_merger(ngroups_tot),stat=err)
  allocate (merged_with   (ngroups_tot),stat=err)
  allocate (majorz        (ngroups_tot),stat=err)
  allocate (t_acc         (ngroups_tot),stat=err)
  allocate (t_appear      (ngroups_tot),stat=err)
  allocate (point         (ngroups_tot),stat=err)
  allocate (bottom        (ngroups_tot),stat=err)
  allocate (name          (ngroups_tot),stat=err)
  allocate (good          (ngroups_tot),stat=err)
#ifdef PLC
  if (Flimit_plc>0.) then
     allocate (Flast         (ngroups_tot),stat=err)
     allocate (zplc          (NPLC_max)   ,stat=err)
     allocate (Mplc          (NPLC_max)   ,stat=err)
     allocate (xplc        (3,NPLC_max)   ,stat=err)
     allocate (vplc        (3,NPLC_max)   ,stat=err)
     allocate (nameplc       (NPLC_max)   ,stat=err)
     total_allocated = total_allocated + 9*8*NPLC_max
  endif
#endif

  if (alloc_verbose) call message ('group vectors allocated')

  write(info,"('Total allocated memory:',f8.3,' Gbyte; max allocated memory: ',f8.3,' Gbyte')") &
       total_allocated/Gbyte,max_allocated/Gbyte
  call message (trim(info))

  return
end subroutine fragment_allocate_group_vectors


subroutine set_obj(grp,F,myobj)

  use tocompute_pos
  use fragment_group_vectors
  implicit none
  integer(kind=singleI),intent(in)::grp
  real(kind=doubleR),intent(in)::F
  type(unit_obj),intent(out)::myobj

  myobj%F=F
  myobj%M=mass(grp)
  myobj%q=centres(:,grp)
  myobj%v=velgrp(:,grp)

  return
end subroutine set_obj


subroutine set_point(i,j,k,ind,F,myobj)

  use fragment_matrices
  use fragment_group_vectors
  use tocompute_pos
  implicit none

  integer(kind=singleI),intent(in)::i,j,k,ind
  real(kind=doubleR),intent(in)::F
  type(unit_obj),intent(out)::myobj

  myobj%F=F
  myobj%M=1
  myobj%q(1)=i-0.5
  myobj%q(2)=j-0.5
  myobj%q(3)=k-0.5
  myobj%v(1)=vx(ind)
  myobj%v(2)=vy(ind)
  myobj%v(3)=vz(ind)

  return
end subroutine set_point


subroutine set_group(grp,myobj)

  use tocompute_pos
  use fragment_group_vectors
  implicit none
  integer(kind=singleI),intent(in)::grp
  type(unit_obj),intent(in)::myobj

  mass(grp)      = myobj%M
  centres(:,grp) = myobj%q
  velgrp(:,grp)  = myobj%v

  return
end subroutine set_group


function q2x(i,myobj)

  use pinocchio_module
  use tocompute_pos
  implicit none
  real(kind=doubleR)::q2x
  integer(kind=singleI),intent(in)::i
  type(unit_obj),intent(in)::myobj

  q2x = myobj%q(i) + myobj%v(i)/myobj%F

  if (pbc_flag(i)) then
     if (q2x>Lgrid(i)) q2x=q2x-dble(Lgrid(i))
     if (q2x<0.)       q2x=q2x+dble(Lgrid(i))
  endif

  return
end function q2x


function vel(i,myobj)

! it gives the velocity of a group in km/s

  use tocompute_pos
  use pinocchio_module
  implicit none
  real(kind=doubleR)::vel
  integer(kind=singleI),intent(in)::i
  type(unit_obj),intent(in)::myobj
  real(kind=doubleR)::z
  real(kind=doubleR),external::fomega,Hubble,inverse_grow

  z=inverse_grow(1./myobj%F)

  vel = myobj%v(i) * Hubble(z)/(1.+z)*InterPartDist*fomega(z) / myobj%F

  return
end function vel


function distance(i)

  use pinocchio_module
  use tocompute_pos
  real(kind=doubleR)::distance
  integer(kind=singleI),intent(in)::i
  real(kind=doubleR),external::q2x

  distance=q2x(i,obj2)-q2x(i,obj1)

  if (pbc_flag(i)) then
     if (distance >  dble(Lgrid(i))/2.) distance = distance - dble(Lgrid(i))
     if (distance < -dble(Lgrid(i))/2.) distance = distance + dble(Lgrid(i))
  endif

end function distance


subroutine accr_cond(i,j,k,ind,Fmax,grp,dd,rr)

! Condition for accretion

  use pinocchio_module
  use fragment_grids
  use fragment_group_vectors
  use tocompute_pos
  implicit none

  integer(kind=singleI),intent(in)::i,j,k,grp,ind
  real(kind=doubleR),intent(in)::Fmax
  real(kind=doubleR),intent(out)::dd,rr

  real(kind=doubleR)::rvir,dx,dy,dz,d2
  real(kind=doubleR),external::virial,distance

  rvir=virial(mass(grp),Fmax,1)
  rr=rvir*rvir
  dd=100.*rr

  call set_point(i,j,k,ind,Fmax,obj1)
  call set_obj(grp,Fmax,obj2)

  dx=distance(1)
  if (abs(dx).lt.rvir) then
     dy=distance(2)
     d2=dx*dx+dy*dy
     if (d2.lt.rr) then
        dz=distance(3)
        d2=d2+dz*dz
        if (d2.le.rr) dd=d2
     endif
  endif

!  write(130+proc_id,'(f12.5,4(1x,i6),10(1x,f12.5))')Fmax,i,j,k,grp,rvir,dx,dd,rr

  return
end subroutine accr_cond


subroutine merge_cond(Fmax,grp1,grp2,merge_flag)

! Checks whether two groups must be merged

  use pinocchio_module
  use fragment_grids
  use fragment_group_vectors
  use tocompute_pos
  implicit none
  real(kind=doubleR),intent(in)::Fmax
  integer(kind=singleI),intent(in)::grp1,grp2
  logical,intent(out)::merge_flag

  real(kind=doubleR)::rvir,rvir1,rvir2,dd,rr,dx,dy,dz
  real(kind=doubleR),external::virial,distance

  merge_flag=.false.
  rvir1=virial(mass(grp1),Fmax,0)
  rvir2=virial(mass(grp2),Fmax,0)
  rvir=max(rvir1,rvir2)
  rr=rvir*rvir

  call set_obj(grp1,Fmax,obj1)
  call set_obj(grp2,Fmax,obj2)

  dx=distance(1)
  if (abs(dx).lt.rvir) then
     dy=distance(2)
     dd=dx*dx+dy*dy
     if (dd.lt.rr) then
        dz=distance(3)
        dd=dd+dz*dz
        if (dd.le.rr) merge_flag=.true.
     endif
  endif

! Bye!

  return
end subroutine merge_cond


subroutine update()

! Updates the centre position of group 1

  use pinocchio_module
  use tocompute_pos
  implicit none

  real(kind=doubleR),external::distance
  real(kind=doubleR)::d
  integer(kind=singleI)::i

  do i=1,3

     d = abs(obj1%q(i) - obj2%q(i))
     if (.not.pbc_flag(i) .or.  d <= dble(Lgrid(i))/2.) then
        obj1%q(i) = (obj1%q(i)*obj1%M + obj2%q(i)*obj2%M)/float(obj1%M + obj2%M)
     else
! PBC must be considered
        if (obj1%q(i) > Lgrid(i)/2) then
! in this case xc is in the second half, and obj2%q(i) -> obj2%q(i)+Lgrid
           obj1%q(i) = (obj1%q(i)*obj1%M + (obj2%q(i)+Lgrid(i))*obj2%M)/float(obj1%M + obj2%M)
        else
! in this case xc is in the first half, and obj2%q(i) -> obj2%q(i)-Lgrid
           obj1%q(i) = (obj1%q(i)*obj1%M + (obj2%q(i)-Lgrid(i))*obj2%M)/float(obj1%M + obj2%M)
        endif
! checks PBC again
        if (obj1%q(i)>Lgrid(i)) obj1%q(i)=obj1%q(i)-Lgrid(i)
        if (obj1%q(i)<0)        obj1%q(i)=obj1%q(i)+Lgrid(i)
     endif

! velocity
     obj1%v(i) = (obj1%v(i)*obj1%M + obj2%v(i)*obj2%M)/float(obj1%M + obj2%M)

  enddo

  obj1%M=obj1%M+obj2%M

! bye!
  return
end subroutine update
