!*****************************************************************
!*                       PINOCCHI0  V3.0                         *
!*  (PINpointing Orbit-Crossing Collapsed HIerarchical Objects)  *
!*****************************************************************
!
!This code was written by
!Pierluigi Monaco, Tom Theuns, Giuliano Taffoni
!Copyright (C) 2013
!
!web page: http//adlibitum.oats.inaf.it/monaco/Homepage/Pinocchio/
!
!This program is free software; you can redistribute it and/or modify
!it under the terms of the GNU General Public License as published by
!the Free Software Foundation; either version 2 of the License, or
!(at your option) any later version.
!
!This program is distributed in the hope that it will be useful,
!but WITHOUT ANY WARRANTY; without even the implied warranty of
!MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!GNU General Public License for more details.
!
!You should have received a copy of the GNU General Public License
!along with this program; if not, write to the Free Software
!Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

module pinocchio_module

  use numbers

! MPI stuff
  include 'mpif.h'
  integer(kind=singleI)  :: ierr, k_offset, proc_id, nproc
  integer(kind=singleI)  :: MPI_STATUS(MPI_STATUS_SIZE)
  real(kind=doubleR)     :: cputime(0:12)=0.0,cputmp

! Cosmological and box parameters
  real(kind=doubleR):: Omega0
  real(kind=doubleR):: OmegaLambda
  real(kind=doubleR):: Hubble100
  real(kind=doubleR):: Sigma8
  real(kind=doubleR):: Omegab
  real(kind=doubleR):: DEw0,DEw1      ! parameters for the dark energy EoS
  real(kind=doubleR):: PowerSpectrum  ! flag for the transfer function
  real(kind=doubleR):: csp            ! normalization constant for P(k)
  real(kind=doubleR):: delta_c        ! delta_c Press & Schechter parameter
  integer(kind=singleI):: filter      ! flag for window filter
  integer(kind=singleI):: flag_ps     ! flag for the analytical mass function
! Basic properties of the highest resolution box
  real(kind=doubleR):: InterPartDist  ! comoving inter-particle distance
  real(kind=doubleR):: BoxSize(3)     ! 3D comoving size of the global box
  real(kind=doubleR):: Mpart          ! particle mass
  real(kind=doubleR)::rho             ! in Msun/Mpc^3
  integer(kind=singleI):: safe        ! border to replicate for each fragmented sub-box
!
  character(len=30)  :: run           ! name of the run
  character(len=*),parameter::dir='Data/'
  character(len=300)::TabulatedEoSfile

  real(kind=doubleR)::SafetyBorder=30.0  ! Boudary layer for fragmentation

! for fragment
  integer(kind=singleI)::Lgrid_frag(3),Npart_frag,nstart(3)

  character(len=300)::Paramfile,OutputList !,StartFlag , ReadDensityField

! logical flags
  logical::WriteRmax
  logical::WriteFmax
  logical::WriteVmax
  logical::WriteDensity
  logical::WriteDisplacements
  !logical::StartFromSavedFmax
  logical::CatalogInAscii
  logical::DoNotFragment
  logical::OutputInH100
  integer(kind=singleI)::RandomSeed, ResamplingRate, saved_seed

! outputs
  integer(kind=singleI)::nout
  real(kind=doubleR),allocatable,dimension(:)::fout,zout

! allocations
  real(kind=doubleR)::total_allocated=0.0,max_allocated=0.0
  logical::alloc_verbose=.false.
!  logical::alloc_verbose=.true.
  real(kind=doubleR), parameter :: Gbyte=1024.d0*1024.d0*1024.d0

! 3D lengths of the global grids (number of particles or grid vertices per side)
  integer(kind=singleI),dimension(3) :: Lgrid
! flag for PBCs on each dimension
  logical,dimension(3)               :: pbc_flag
! Number of sub-boxes per dimension
  integer(kind=singleI),dimension(3) :: nbox

! string for i/o
  character(len=300) :: info

! ram requirement
  real(kind=doubleR)::bytes_per_particle=110.

! Past Light Cone
  real(kind=doubleR)::StartingzForPLC
#ifdef PLC
  integer(kind=singleI)::number_of_rep
  real(kind=doubleR)::Flimit_plc
  integer(kind=singleI)::NPLC_max,NPLC=0
#endif

  integer(kind=singleI),parameter::MIN_MASS_FOR_CAT=100

! constants
  real(kind=doubleR), parameter :: pi   =3.1415926535897931d0
  real(kind=doubleR), parameter :: twopi=2.d0*pi

end module pinocchio_module
